from typing import Iterator

from dissect.cim import cim

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export, internal

ConsumerBindingRecord = TargetRecordDescriptor(
    "filesystem/windows/cim/consumerbinding",
    [
        ("string", "query"),
    ],
)


class CimPlugin(Plugin):
    """CIM database plugin.

    Provides functions for getting useful data out the CIM (WBEM) database.
    """

    __namespace__ = "cim"

    def __init__(self, target):
        super().__init__(target)
        self._repo = None

        repodir = self.target.fs.path("sysvol/windows/system32/wbem/repository")
        if repodir.exists():
            index = repodir.joinpath("index.btr")
            objects = repodir.joinpath("objects.data")
            mappings = [repodir.joinpath(f"mapping{i}.map") for i in range(1, 4)]

            if all([index.exists(), objects.exists(), all([m.exists() for m in mappings])]):
                try:
                    self._repo = cim.CIM(index.open(), objects.open(), [m.open() for m in mappings])
                except cim.Error as e:
                    self.target.log.warning("Error opening CIM database", exc_info=e)

    def check_compatible(self) -> None:
        if not self._repo:
            raise UnsupportedPluginError("No WBEM repository found")

    @internal
    def repo(self) -> cim.CIM:
        return self._repo

    @export(record=ConsumerBindingRecord)
    def consumerbindings(self) -> Iterator[ConsumerBindingRecord]:
        """Return all __FilterToConsumerBinding queries.

        WMI permanent event subscriptions can be used to trigger actions when specified conditions are met. Attackers
        often use this functionality to persist the execution of backdoors at system start up. WMI Consumers specify an
        action to be performed, including executing a command, running a script, adding an entry to a log, or sending
        an email. WMI Filters define conditions that will trigger a Consumer.

        References:
            - https://learn-powershell.net/2013/08/14/powershell-and-events-permanent-wmi-event-subscriptions/
            - https://www.mandiant.com/resources/dissecting-one-ofap
            - https://support.sophos.com/support/s/article/KB-000038535?language=en_US&c__displayLanguage=en_US
        """
        subscription_ns = self._repo.root.namespace("subscription")
        try:
            for binding in subscription_ns.class_("__filtertoconsumerbinding").instances:
                consumer = subscription_ns.query(binding.properties["Consumer"].value)
                if query := consumer.properties.get("CommandLineTemplate"):
                    yield ConsumerBindingRecord(
                        query=query.value,
                        _target=self.target,
                    )
        except Exception as e:  # noqa
            self.target.log.warning("Error during consumerbindings execution", exc_info=e)
            pass



import struct
from typing import Iterator

from dissect.util.ts import wintimestamp

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.regf.shellbags import (
    FILE_ENTRY,
    parse_shell_item_list,
)

UserRegistryRecordDescriptor = create_extended_descriptor(
    [
        RegistryRecordDescriptorExtension,
        UserRecordDescriptorExtension,
    ]
)

RunMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/run",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

RecentDocsRecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/recentdocs",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

OpenSaveMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/opensave",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

LastVisitedMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/lastvisited",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "filename"),
        ("string", "path"),
        ("string", "key"),
    ],
)

ACMruRecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/acmru",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "category"),
        ("string", "value"),
        ("string", "key"),
    ],
)

MapNetworkDriveMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/mapnetworkdrive",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

TerminalServerMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/mstsc",
    [
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)

MSOfficeMRURecord = UserRegistryRecordDescriptor(
    "windows/registry/mru/msoffice",
    [
        ("datetime", "ts"),
        ("datetime", "regf_mtime"),
        ("varint", "index"),
        ("string", "value"),
        ("string", "key"),
    ],
)


class MRUPlugin(Plugin):
    """Return MRU data stored at various registry keys.

    The Windows registry contains various keys about Most Recently Used (MRU) files.

    References:
        - https://winreg-kb.readthedocs.io/en/latest/sources/explorer-keys/Most-recently-used.html
    """

    __namespace__ = "mru"

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            return UnsupportedPluginError("Target has no registry")

    @export(record=RunMRURecord)
    def run(self) -> Iterator[RunMRURecord]:
        """Return the RunMRU data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RunMRU`` registry key contains information
        about the most recent commands that have been performed by the Run application

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """
        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RunMRU"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_key(self.target, key, RunMRURecord)

    @export(record=RecentDocsRecord)
    def recentdocs(self) -> Iterator[RecentDocsRecord]:
        """Return the RecentDocs data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RecentDocs`` registry key contains
        information about the last 10 documents that the currently logged on user accessed or executed via Windows
        Explorer.

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RecentDocs"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_ex_key(self.target, key, RecentDocsRecord)

    @export(record=OpenSaveMRURecord)
    def opensave(self) -> Iterator[OpenSaveMRURecord]:
        """Return the OpenSaveMRU data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\OpenSaveMRU`` registry key
        contains information about the most recently opened or saved files.

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\OpenSaveMRU"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_key(self.target, key, OpenSaveMRURecord)

    @export(record=LastVisitedMRURecord)
    def lastvisited(self) -> Iterator[LastVisitedMRURecord]:
        """Return the LastVisitedMRU data.

        The ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedMRU`` registry key
        contains information about the executable used by an application to open the files that are documented at the
        OpenSaveMRU registry key. Also each value tracks the directory location for the last file that was accessed by
        that application.

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/
        """

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedMRU"

        for key in self.target.registry.keys(KEY):
            user = self.target.registry.get_user(key)

            try:
                mrulist = key.value("MRUList").value
            except RegistryError:
                mrulist = None

            for value in key.values():
                if value.name == "MRUList":
                    continue

                entry_index = mrulist.index(value.name) if mrulist else None
                filename, path, _ = value.value.rsplit(b"\x00\x00")

                yield LastVisitedMRURecord(
                    regf_mtime=key.ts,
                    index=entry_index,
                    filename=filename.decode("utf-16-le"),
                    path=path.decode("utf-16-le"),
                    key=key.path,
                    _target=self.target,
                    _user=user,
                    _key=key,
                )

    @export(record=ACMruRecord)
    def acmru(self) -> Iterator[ACMruRecord]:
        """Return the ACMru (Windows Search) data.

        The following keys are being searched:
          - ``HKCU\\Software\\Microsoft\\Search Assistant\\ACMru``:
            This registry key contains the most recent search history from Windows default search.
          - ``HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\WordWheelQuery``:
            This registry key contains the most recent search history from Windows Explorer. (Windows >=7)

        References:
            - https://digitalf0rensics.wordpress.com/2014/01/17/windows-registry-and-forensics-part2/

        Known categories:
            - 5001: Internet Search Assistant
            - 5603: Windows XP files and folder search
            - 5604: "Word or phrase in a file" dialog box
            - 5647: "For computers or people" selection in Search Results dialog box
        """

        KEY = "HKCU\\Software\\Microsoft\\Search Assistant\\ACMru"

        for key in self.target.registry.keys(KEY):
            user = self.target.registry.get_user(key)
            for subkey in key.subkeys():
                for value in subkey.values():
                    yield ACMruRecord(
                        regf_mtime=key.ts,
                        index=int(value.name),
                        category=subkey.name,
                        value=value.value,
                        key=key.path,
                        _target=self.target,
                        _user=user,
                        _key=subkey,
                    )

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\WordWheelQuery"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_ex_key(self.target, key, ACMruRecord)

    @export(record=MapNetworkDriveMRURecord)
    def networkdrive(self) -> Iterator[MapNetworkDriveMRURecord]:
        """Return MRU of mapped network drives.

        The HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Map Network Drive MRU registry key contains
        information about the most recently used mapped network drives.

        References:
            - https://winreg-kb.readthedocs.io/en/latest/sources/explorer-keys/Most-recently-used.html#keys-with-a-mrulist-value
        """  # noqa: E501

        KEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Map Network Drive MRU"

        for key in self.target.registry.keys(KEY):
            yield from parse_mru_key(self.target, key, MapNetworkDriveMRURecord)

    @export(record=TerminalServerMRURecord)
    def mstsc(self) -> Iterator[TerminalServerMRURecord]:
        """Return Terminal Server Client MRU data."""

        KEY = "HKCU\\Software\\Microsoft\\Terminal Server Client\\Default"

        for key in self.target.registry.keys(KEY):
            user = self.target.registry.get_user(key)

            for value in key.values():
                entry_index = int(value.name.split("MRU")[1])

                yield TerminalServerMRURecord(
                    regf_mtime=key.ts,
                    index=entry_index,
                    value=value.value,
                    key=key.path,
                    _target=self.target,
                    _user=user,
                    _key=key,
                )

    @export(record=MSOfficeMRURecord)
    def msoffice(self) -> Iterator[MSOfficeMRURecord]:
        """Return MS Office MRU keys."""

        KEY = "HKCU\\Software\\Microsoft\\Office"
        SUBKEYS = [
            "Common",
            "Excel",
            "Groove",
            "OneNote",
            "Outlook",
            "PowerPoint",
            "Publisher",
            "Word",
        ]

        for key in self.target.registry.keys(KEY):
            for version_key in key.subkeys():
                if not version_key.name[0].isdigit():
                    continue

                for subkey in version_key.subkeys():
                    if subkey.name not in SUBKEYS:
                        continue

                    try:
                        yield from parse_office_mru(self.target, subkey, MSOfficeMRURecord)
                    except RegistryError:
                        pass


def parse_mru_key(target, key, record):
    user = target.registry.get_user(key)

    try:
        mrulist = key.value("MRUList").value
    except RegistryError:
        mrulist = None

    for value in key.values():
        if value.name == "MRUList":
            continue

        entry_index = mrulist.index(value.name) if mrulist else None
        entry_value = value.value

        yield record(
            regf_mtime=key.ts,
            index=entry_index,
            value=entry_value,
            key=key.path,
            _target=target,
            _user=user,
            _key=key,
        )

    for subkey in key.subkeys():
        yield from parse_mru_key(target, subkey, record)


def parse_mru_ex_key(target, key, record):
    user = target.registry.get_user(key)

    mrulist_ex = key.value("MRUListEx").value
    mrulist_ex = struct.unpack(f"<{len(mrulist_ex) // 4}I", mrulist_ex)

    for value in key.values():
        if value.name == "MRUListEx":
            continue

        entry_index = mrulist_ex.index(int(value.name))
        split_idx = value.value.index(b"\x00\x00")
        # Poor mans null terminated utf-16-le
        path, bag = value.value[: split_idx + 1], value.value[split_idx + 3 :]
        parsed_bag = list(parse_shell_item_list(bag))
        if len(parsed_bag) != 1 or not isinstance(parsed_bag, FILE_ENTRY):
            target.log.debug("Unexpected shell bag entry in MRUListEx entry: %s:%s", key, value)

        yield record(
            regf_mtime=key.ts,
            index=entry_index,
            value=path.decode("utf-16-le"),
            key=key.path,
            _target=target,
            _user=user,
            _key=key,
        )


def parse_office_mru(target, key, record):
    try:
        yield from parse_office_mru_key(target, key.subkey("File MRU"), record)
    except RegistryError:
        pass

    try:
        yield from parse_office_mru_key(target, key.subkey("Place MRU"), record)
    except RegistryError:
        pass

    try:
        for subkey in key.subkey("User MRU").subkeys():
            yield from parse_office_mru(target, subkey, record)
    except RegistryError:
        pass


def parse_office_mru_key(target, key, record):
    user = target.registry.get_user(key)

    for value in key.values():
        if not value.name.startswith("Item"):
            continue

        entry_index = int(value.name.split(" ")[1])
        info_str, path = value.value.split("*", 1)

        info = {part[0]: int(part[1:], 16) for part in info_str.strip("[]").split("][")}

        yield record(
            ts=wintimestamp(info["T"]),
            regf_mtime=key.ts,
            index=entry_index,
            value=path,
            key=key.path,
            _target=target,
            _user=user,
            _key=key,
        )



import hashlib
import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import BinaryIO, Iterator, Optional
from uuid import UUID

from dissect.cstruct import cstruct
from dissect.util.sid import read_sid

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers import keychain
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.general.users import UserDetails
from dissect.target.plugins.os.windows.dpapi.crypto import (
    CipherAlgorithm,
    HashAlgorithm,
    derive_password_hash,
)
from dissect.target.target import Target

log = logging.getLogger(__name__)


CredHistRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/credential/history",
    [
        ("string", "guid"),
        ("boolean", "decrypted"),
        ("string", "sha1"),
        ("string", "nt"),
    ],
)


credhist_def = """
struct entry {
    DWORD   dwVersion;
    CHAR    guidLink[16];
    DWORD   dwNextLinkSize;
    DWORD   dwCredLinkType;
    DWORD   algHash;                    // ALG_ID
    DWORD   dwPbkdf2IterationCount;
    DWORD   dwSidSize;
    DWORD   algCrypt;                   // ALG_ID
    DWORD   dwShaHashSize;
    DWORD   dwNtHashSize;
    CHAR    pSalt[16];
    CHAR    pSid[dwSidSize];
    CHAR    encrypted[0];
};
"""

c_credhist = cstruct().load(credhist_def)


@dataclass
class CredHistEntry:
    version: int
    guid: str
    user_sid: str
    sha1: Optional[bytes]
    nt: Optional[bytes]
    hash_alg: HashAlgorithm = field(repr=False)
    cipher_alg: CipherAlgorithm = field(repr=False)
    raw: c_credhist.entry = field(repr=False)
    decrypted: bool = False

    def decrypt(self, password_hash: bytes) -> None:
        """Decrypt this CREDHIST entry using the provided password hash. Modifies ``CredHistEntry.sha1``
        and ``CredHistEntry.nt`` values.

        If the decrypted ``nt`` value is 16 bytes we assume the decryption was successful.

        Args:
            password_hash: Bytes of SHA1 password hash digest.

        Raises:
            ValueError: If the decryption seems to have failed.
        """
        data = self.cipher_alg.decrypt_with_hmac(
            data=self.raw.encrypted,
            key=derive_password_hash(password_hash, self.user_sid),
            iv=self.raw.pSalt,
            hash_algorithm=self.hash_alg,
            rounds=self.raw.dwPbkdf2IterationCount,
        )

        sha_size = self.raw.dwShaHashSize
        nt_size = self.raw.dwNtHashSize

        sha1 = data[:sha_size]
        nt = data[sha_size : sha_size + nt_size].rstrip(b"\x00")

        if len(nt) != 16:
            raise ValueError("Decrypting failed, invalid password hash?")

        self.decrypted = True
        self.sha1 = sha1
        self.nt = nt


class CredHistFile:
    def __init__(self, fh: BinaryIO) -> None:
        self.fh = fh
        self.entries = list(self._parse())

    def __repr__(self) -> str:
        return f"<CredHistFile fh='{self.fh}' entries={len(self.entries)}>"

    def _parse(self) -> Iterator[CredHistEntry]:
        self.fh.seek(0)
        try:
            while True:
                entry = c_credhist.entry(self.fh)

                # determine size of encrypted data and add to entry
                cipher_alg = CipherAlgorithm.from_id(entry.algCrypt)
                enc_size = entry.dwShaHashSize + entry.dwNtHashSize
                enc_size += enc_size % cipher_alg.block_length
                entry.encrypted = self.fh.read(enc_size)

                yield CredHistEntry(
                    version=entry.dwVersion,
                    guid=UUID(bytes_le=entry.guidLink),
                    user_sid=read_sid(entry.pSid),
                    hash_alg=HashAlgorithm.from_id(entry.algHash),
                    cipher_alg=cipher_alg,
                    sha1=None,
                    nt=None,
                    raw=entry,
                )
        except EOFError:
            pass

    def decrypt(self, password_hash: bytes) -> None:
        """Decrypt a CREDHIST chain using the provided password SHA1 hash."""

        for entry in reversed(self.entries):
            try:
                entry.decrypt(password_hash)
            except ValueError as e:
                log.warning("Could not decrypt entry %s with password %s", entry.guid, password_hash.hex())
                log.debug("", exc_info=e)
                continue
            password_hash = entry.sha1


class CredHistPlugin(Plugin):
    """Windows CREDHIST file parser.

    Windows XP:         ``C:\\Documents and Settings\\username\\Application Data\\Microsoft\\Protect\\CREDHIST``
    Windows 7 and up:   ``C:\\Users\\username\\AppData\\Roaming\\Microsoft\\Protect\\CREDHIST``

    Resources:
        - https://www.passcape.com/index.php?section=docsys&cmd=details&id=28#41
    """

    def __init__(self, target: Target):
        super().__init__(target)
        self.files = list(self._find_files())

    def _find_files(self) -> Iterator[tuple[UserDetails, Path]]:
        hashes = set()
        for user_details in self.target.user_details.all_with_home():
            for path in ["AppData/Roaming/Microsoft/Protect/CREDHIST", "Application Data/Microsoft/Protect/CREDHIST"]:
                credhist_path = user_details.home_path.joinpath(path)
                if credhist_path.exists() and (hash := credhist_path.get().hash()) not in hashes:
                    hashes.add(hash)
                    yield user_details.user, credhist_path

    def check_compatible(self) -> None:
        if not self.files:
            raise UnsupportedPluginError("No CREDHIST files found on target.")

    @export(record=CredHistRecord)
    def credhist(self) -> Iterator[CredHistRecord]:
        """Yield and decrypt all Windows CREDHIST entries on the target."""
        passwords = keychain_passwords()

        if not passwords:
            self.target.log.warning("No passwords provided in keychain, cannot decrypt CREDHIST hashes")

        for user, path in self.files:
            credhist = CredHistFile(path.open("rb"))

            for password in passwords:
                credhist.decrypt(hashlib.sha1(password.encode("utf-16-le")).digest())

            for entry in credhist.entries:
                yield CredHistRecord(
                    guid=entry.guid,
                    decrypted=entry.decrypted,
                    sha1=entry.sha1.hex() if entry.sha1 else None,
                    nt=entry.nt.hex() if entry.nt else None,
                    _user=user,
                    _target=self.target,
                )


def keychain_passwords() -> set:
    passphrases = set()
    for key in keychain.get_keys_for_provider("user") + keychain.get_keys_without_provider():
        if key.key_type == keychain.KeyType.PASSPHRASE:
            passphrases.add(key.value)
    passphrases.add("")
    return passphrases



from __future__ import annotations

import gzip
import io
import logging
import zlib
from itertools import cycle, islice
from pathlib import Path
from typing import BinaryIO

from dissect.util.stream import AlignedStream, RangeStream, RelativeStream

from dissect.target.container import Container
from dissect.target.tools.utils import catch_sigpipe

log = logging.getLogger(__name__)


def find_xor_key(fh: BinaryIO) -> bytes:
    """Find the XOR key for the firmware file by using known plaintext of zeros.

    File-like object ``fh`` must be seeked to the correct offset where it should decode to all zeroes (0x00).

    Arguments:
        fh: File-like object to read from.

    Returns:
        bytes: XOR key, note that the XOR key is not validated and may be incorrect.
    """
    key = bytearray()

    pos = fh.tell()
    buf = fh.read(32)
    fh.seek(pos)

    if pos % 512 == 0:
        xor_char = 0xFF
    else:
        fh.seek(pos - 1)
        xor_char = ord(fh.read(1))

    for i, k_char in enumerate(buf):
        idx = (i + pos) & 0x1F
        key.append((xor_char ^ k_char ^ idx) & 0xFF)
        xor_char = k_char

    # align xor key
    koffset = 32 - (pos & 0x1F)
    key = islice(cycle(key), koffset, koffset + 32)
    return bytes(key)


class FortiFirmwareFile(AlignedStream):
    """Fortinet firmware file, handles transparant decompression and deobfuscation of the firmware file."""

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.trailer_offset = None
        self.trailer_data = None
        self.xor_key = None
        self.is_gzipped = False

        size = None

        # Check if the file is gzipped
        self.fh.seek(0)
        header = self.fh.read(4)
        if header.startswith(b"\x1f\x8b"):
            self.is_gzipped = True

            # Find the extra metadata behind the gzip compressed data
            # as a bonus we can also calculate the size of the firmware here
            dec = zlib.decompressobj(wbits=16 + zlib.MAX_WBITS)
            self.fh.seek(0)
            size = 0
            while True:
                data = self.fh.read(io.DEFAULT_BUFFER_SIZE)
                if not data:
                    break
                d = dec.decompress(dec.unconsumed_tail + data)
                size += len(d)

            # Ignore the trailer data of the gzip file if we have any
            if dec.unused_data:
                self.trailer_offset = self.fh.seek(-len(dec.unused_data), io.SEEK_END)
                self.trailer_data = self.fh.read()
                log.info("Found trailer offset: %d, data: %r", self.trailer_offset, self.trailer_data)
                self.fh = RangeStream(self.fh, 0, self.trailer_offset)

            self.fh.seek(0)
            self.fh = gzip.GzipFile(fileobj=self.fh)

        # Find the xor key based on known offsets where the firmware should decode to zero bytes
        for zero_offset in (0x30, 0x40, 0x400):
            self.fh.seek(zero_offset)
            xor_key = find_xor_key(self.fh)
            if xor_key.isalnum():
                self.xor_key = xor_key
                log.info("Found key %r @ offset %s", self.xor_key, zero_offset)
                break
        else:
            log.info("No xor key found")

        # Determine the size of the firmware file if we didn't calculate it yet
        if size is None:
            size = self.fh.seek(0, io.SEEK_END)

        log.info("firmware size: %s", size)
        log.info("xor key: %r", self.xor_key)
        log.info("gzipped: %s", self.is_gzipped)
        self.fh.seek(0)

        # Align the stream to 512 bytes which simplifies the XOR deobfuscation code
        super().__init__(size=size, align=512)

    def _read(self, offset: int, length: int) -> bytes:
        self.fh.seek(offset)
        buf = self.fh.read(length)

        if not self.xor_key:
            return buf

        buf = bytearray(buf)
        xor_char = 0xFF
        for i, cur_char in enumerate(buf):
            if (i + offset) % 512 == 0:
                xor_char = 0xFF
            idx = (i + offset) & 0x1F
            buf[i] = ((self.xor_key[idx] ^ cur_char ^ xor_char) - idx) & 0xFF
            xor_char = cur_char

        return bytes(buf)


class FortiFirmwareContainer(Container):
    __type__ = "fortifw"

    def __init__(self, fh: BinaryIO | Path, *args, **kwargs):
        if not hasattr(fh, "read"):
            fh = fh.open("rb")

        # Open the firmware file
        self.ff = FortiFirmwareFile(fh)

        # seek to MBR
        self.fw = RelativeStream(self.ff, 0x200)
        super().__init__(self.fw, self.ff.size, *args, **kwargs)

    @staticmethod
    def detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return False

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        # all Fortinet firmware files end with `-FORTINET.out`
        return str(path).lower().endswith("-fortinet.out")

    def read(self, length: int) -> bytes:
        return self.fw.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.fw.seek(offset, whence)

    def tell(self) -> int:
        return self.fw.tell()

    def close(self) -> None:
        pass


@catch_sigpipe
def main(argv: list[str] | None = None) -> None:
    import argparse
    import shutil
    import sys

    parser = argparse.ArgumentParser(description="Decompress and deobfuscate Fortinet firmware file to stdout.")
    parser.add_argument("file", type=argparse.FileType("rb"), help="Fortinet firmware file")
    parser.add_argument("--verbose", "-v", action="store_true", help="verbose output")
    args = parser.parse_args(argv)

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    ff = FortiFirmwareFile(args.file)
    shutil.copyfileobj(ff, sys.stdout.buffer)


if __name__ == "__main__":
    main()



from typing import Iterator, Optional

import pytest
from dissect.util.ts import webkittimestamp
from flow.record.fieldtypes import datetime as dt

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.helpers import keychain
from dissect.target.helpers.regutil import VirtualHive
from dissect.target.plugins.apps.browser.chrome import ChromePlugin
from dissect.target.plugins.os.windows.dpapi.dpapi import DPAPIPlugin
from tests._utils import absolute_path
from tests.conftest import add_win_user
from tests.plugins.os.windows.credential.test_lsa import (
    POLICY_KEY_PATH_NT6,
    map_lsa_polkey,
    map_lsa_secrets,
    map_lsa_system_keys,
)
from tests.plugins.os.windows.test__os import map_version_value


@pytest.fixture
def target_chrome_win(target_win_users: Target, fs_win: VirtualFilesystem) -> Iterator[Target]:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Default\\",
        absolute_path("_data/plugins/apps/browser/chrome/"),
    )

    target_win_users.add_plugin(ChromePlugin)

    yield target_win_users


@pytest.fixture
def target_chrome_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Iterator[Target]:
    fs_unix.map_dir("/root/.config/google-chrome/Default/", absolute_path("_data/plugins/apps/browser/chrome/"))

    target_unix_users.add_plugin(ChromePlugin)

    yield target_unix_users


@pytest.fixture
def target_chrome_win_snapshot(target_win_users: Target, fs_win: VirtualFilesystem) -> Iterator[Target]:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\116.0.5038.150\\Default",
        absolute_path("_data/plugins/apps/browser/chrome/"),
    )

    target_win_users.add_plugin(ChromePlugin)

    yield target_win_users


@pytest.mark.parametrize(
    "target_platform",
    ["target_chrome_win", "target_chrome_unix", "target_chrome_win_snapshot"],
)
def test_chrome_history(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chrome.history())

    assert len(records) == 5
    assert set(["chrome"]) == set(record.browser for record in records)

    assert (
        records[0].url == "https://www.google.com/search?q=github+fox-it+dissect&oq=github+fox-it+dissect"
        "&aqs=chrome..69i57.12832j0j4&sourceid=chrome&ie=UTF-8"
    )
    assert records[0].id == "1"
    assert records[0].visit_count == 2
    assert records[0].ts == dt("2023-02-24T11:54:07.157810+00:00")


@pytest.mark.parametrize(
    "target_platform",
    ["target_chrome_win", "target_chrome_unix", "target_chrome_win_snapshot"],
)
def test_chrome_downloads(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chrome.downloads())

    assert len(records) == 1
    assert set(["chrome"]) == set(record.browser for record in records)

    assert records[0].id == 6
    assert records[0].ts_start == dt("2023-02-24T11:54:19.726147+00:00")
    assert records[0].ts_end == dt("2023-02-24T11:54:21.030043+00:00")
    assert records[0].url == "https://codeload.github.com/fox-it/dissect/zip/refs/heads/main"


@pytest.mark.parametrize(
    "target_platform",
    ["target_chrome_win", "target_chrome_unix", "target_chrome_win_snapshot"],
)
def test_chrome_extensions(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.chrome.extensions())

    assert len(records) == 8
    assert set(["chrome"]) == set(record.browser for record in records)

    assert records[0].ts_install == dt("2022-11-24T15:20:43.682152+00:00")
    assert records[0].ts_update == dt("2022-11-24T15:20:43.682152+00:00")
    assert records[0].name == "Web Store"
    assert records[0].version == "0.2"
    assert records[0].id == "ahfgeienlihckogmohjhadlkjgocpleb"


def test_windows_chrome_passwords_plugin(target_chrome_win: Target) -> None:
    records = list(target_chrome_win.chrome.passwords())

    assert len(records) == 2

    for record in records:
        assert record.browser == "chrome"
        assert record.decrypted_username == "username"
        assert record.decrypted_password is None

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_chrome_passwords_basic_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir(
        "/root/.config/google-chrome/Default/", absolute_path("_data/plugins/apps/browser/chromium/unix/basic/")
    )
    target_unix_users.add_plugin(ChromePlugin)

    records = list(target_unix_users.chrome.passwords())

    assert len(records) == 2

    for record in records:
        assert record.browser == "chrome"
        assert record.decrypted_username == "username"
        assert record.decrypted_password == "password"

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_chrome_passwords_gnome_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir(
        "/root/.config/google-chrome/Default/", absolute_path("_data/plugins/apps/browser/chromium/unix/gnome/")
    )
    target_unix_users.add_plugin(ChromePlugin)

    records = list(target_unix_users.chrome.passwords())

    assert len(records) == 1

    assert records[0].decrypted_username == "username"
    assert records[0].decrypted_password is None
    assert records[0].url == "https://test.com/"


@pytest.fixture
def target_win_users_dpapi(
    hive_hklm: VirtualHive, hive_hku: VirtualHive, fs_win: VirtualFilesystem, target_win: Target
) -> Iterator[Target]:
    # Add users
    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="%systemroot%\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid="S-1-5-21-1342509979-482553916-3960431919-1000",
        home="C:\\Users\\user",
    )

    # Add system dpapi files
    fs_win.map_dir(
        "Windows/System32/Microsoft/Protect",
        absolute_path("_data/plugins/os/windows/dpapi/fixture/Protect_System32"),
    )

    # Add user dpapi files
    fs_win.map_dir(
        "Users/User/AppData/Roaming/Microsoft/Protect",
        absolute_path("_data/plugins/os/windows/dpapi/fixture/Protect_User"),
    )

    # Add registry dpapi keys
    map_lsa_system_keys(hive_hklm, {"Data": "8fa8e1fb", "GBG": "a6e23eb8", "JD": "fe5ffdaf", "Skew1": "6e289261"})
    map_lsa_polkey(
        hive_hklm,
        POLICY_KEY_PATH_NT6,
        bytes.fromhex(
            "00000001ecffe17b2a997440aa939adbff26f1fc0300000000000000676f5836"
            "37c3e0e7b9edf43b3b29b1d0d24cb6bfc60e0fc4dc446e7d244d303533b90a2b"
            "d732fcf985748a8917aea73e9d0b290ee4ba2f53e6a9a0ac9b3c9b26e721b01b"
            "7a6c1f92b517e2a33f5f6de7f736716793b19872059595e6b4dc888d19d1d615"
            "d602bed553478c411d2fed045602ddbb5adc31c9901021ad339bca368bdb554f"
            "fe074a7074528d5e9dcbb467"
        ),
    )
    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "000000017c3e71eca8fb4eed03ea4361fbc783870300000000000000af64ca32"
                "a15059f8e38f328a5f16d063939bdbb9321ba159dcafd9cdf316d82f89a829d7"
                "58024b276d099ef2290ca46fc7b2635568500bf2d31ed8ce1e0330345cca5ef3"
                "e8d18399a22ae88db12872ee5bb0c1f0dd3b83066269d0d9618b19bb"
            )
        },
    )

    target_win.add_plugin(DPAPIPlugin)
    yield target_win


@pytest.mark.parametrize(
    "keychain_value, expected_password",
    [
        ("user", "StrongPassword"),
        ("invalid", None),
    ],
)
def test_windows_chrome_passwords_dpapi(
    target_win_users_dpapi: Target, fs_win: VirtualFilesystem, keychain_value: str, expected_password: Optional[str]
) -> None:
    fs_win.map_dir(
        "Users/user/AppData/Local/Google/Chrome/User Data",
        absolute_path("_data/plugins/apps/browser/chrome/dpapi/User_Data"),
    )

    map_version_value(target_win_users_dpapi, "CurrentVersion", 10.0)

    target_win_users_dpapi.add_plugin(ChromePlugin)

    keychain.KEYCHAIN.clear()
    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value=keychain_value,
        identifier=None,
        provider="user",
    )

    records = list(target_win_users_dpapi.chrome.passwords())

    assert len(keychain.get_all_keys()) == 1
    assert len(records) == 2

    assert records[0].url == "https://example.com/"
    assert records[0].encrypted_password == "djEwT8fVcC9jiZPrMl8QdcFGSlfNArTPJG7Q/Wz4svHp9cRVG1NqC1/Jc8QR"
    assert records[0].decrypted_password == expected_password


def test_windows_chrome_cookies_dpapi(target_win_users_dpapi: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_dir(
        "Users/user/AppData/Local/Google/Chrome/User Data",
        absolute_path("_data/plugins/apps/browser/chrome/dpapi/User_Data"),
    )

    map_version_value(target_win_users_dpapi, "CurrentVersion", 10.0)

    target_win_users_dpapi.add_plugin(ChromePlugin)

    keychain.KEYCHAIN.clear()
    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="user",
        identifier=None,
        provider="user",
    )

    records = list(target_win_users_dpapi.chrome.cookies())

    assert len(records) == 4

    assert records[0].ts_created == webkittimestamp(13370000000000000)
    assert records[0].ts_last_accessed == webkittimestamp(13370000000000000)
    assert records[0].browser == "chrome"
    assert records[0].name == "tbb"
    assert records[0].value == "false"
    assert records[0].host == ".tweakers.net"
    assert records[0].is_secure

    assert {c.name: c.value for c in records} == {
        "tbb": "false",
        "twk-theme": "twk-light",
        "GPS": "1",
        "PREF": "tz=Europe.Berlin",
    }


def test_chrome_windows_snapshots(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    base_dir = "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Default"
    snapshot_dirs = [
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\116.0.5038.150\\Default",
        "Users\\John\\AppData\\Local\\Google\\Chrome\\User Data\\Snapshots\\119.0.7845.119\\Default",
    ]
    profile_dirs = [base_dir] + snapshot_dirs

    for dir in profile_dirs:
        fs_win.map_dir(
            dir,
            absolute_path("_data/plugins/apps/browser/chrome/"),
        )

    target_win_users.add_plugin(ChromePlugin)

    records_list = [
        list(target_win_users.chrome.history()),
        list(target_win_users.chrome.extensions()),
        list(target_win_users.chrome.downloads()),
    ]

    # Loop over the different types of records and verify we have the same amount of records in each profile directory.
    for records in records_list:
        assert set(["chrome"]) == set(record.browser for record in records)

        base_path_records = [r for r in records if str(r.source.parent).endswith(base_dir)]

        for snapshot_dir in snapshot_dirs:
            # Retrieve records that are in the snapshot's directory.
            snapshot_records = [r for r in records if str(r.source.parent).endswith(snapshot_dir)]

            # We map the same files in each of the snapshot directories.
            assert len(base_path_records) == len(snapshot_records)



import textwrap
from datetime import datetime, timezone
from io import BytesIO

from dissect.target.plugins.apps.webserver.nginx import NginxPlugin
from tests._utils import absolute_path


def test_plugins_apps_webservers_nginx_txt(target_unix, fs_unix):
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log")
    fs_unix.map_file("var/log/nginx/access.log", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[0]
    assert record.ts == datetime(2022, 12, 1, 0, 3, 57, tzinfo=timezone.utc)
    assert record.status_code == 200
    assert record.remote_ip == "1.2.3.4"
    assert record.remote_user == "admin"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_plugins_apps_webservers_nginx_ipv6(target_unix, fs_unix):
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log")
    fs_unix.map_file("var/log/nginx/access.log", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[1]
    assert record.remote_ip == "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    assert record.status_code == 200
    assert record.remote_user == "-"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_plugins_apps_webservers_nginx_gz(target_unix, fs_unix):
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log.gz")
    fs_unix.map_file("var/log/nginx/access.log.1.gz", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[0]
    assert record.status_code == 200
    assert record.remote_ip == "1.2.3.4"
    assert record.remote_user == "admin"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_plugins_apps_webservers_nginx_bz2(target_unix, fs_unix):
    data_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log.bz2")
    fs_unix.map_file("var/log/nginx/access.log.1.bz2", data_file)

    target_unix.add_plugin(NginxPlugin)
    results = list(target_unix.nginx.access())

    assert len(results) == 2

    record = results[0]
    assert record.status_code == 200
    assert record.remote_ip == "1.2.3.4"
    assert record.remote_user == "admin"
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.bytes_sent == 123


def test_plugins_apps_webservers_nginx_config(target_unix, fs_unix):
    config_file = absolute_path("_data/plugins/apps/webserver/nginx/nginx.conf")
    fs_unix.map_file("etc/nginx/nginx.conf", config_file)

    for i, log in enumerate(["access.log", "domain1.access.log", "domain2.access.log", "big.server.access.log"]):
        fs_unix.map_file_fh(f"opt/logs/{i}/{log}", BytesIO(b"Foo"))

    target_unix.add_plugin(NginxPlugin)
    log_paths = target_unix.nginx.get_log_paths()

    assert len(log_paths) == 4


def test_plugins_apps_webservers_nginx_config_logs_logrotated(target_unix, fs_unix):
    config_file = absolute_path("_data/plugins/apps/webserver/nginx/nginx.conf")
    fs_unix.map_file("etc/nginx/nginx.conf", config_file)
    fs_unix.map_file_fh("opt/logs/0/access.log", BytesIO(b"Foo1"))
    fs_unix.map_file_fh("opt/logs/0/access.log.1", BytesIO(b"Foo2"))
    fs_unix.map_file_fh("opt/logs/0/access.log.2", BytesIO(b"Foo3"))
    fs_unix.map_file_fh("opt/logs/1/domain1.access.log", BytesIO(b"Foo4"))
    fs_unix.map_file_fh("var/log/nginx/access.log", BytesIO(b"Foo5"))

    target_unix.add_plugin(NginxPlugin)
    log_paths = target_unix.nginx.get_log_paths()

    assert len(log_paths) == 5


def test_plugins_apps_webservers_nginx_config_commented_logs(target_unix, fs_unix):
    config = """
    # access_log      /foo/bar/old.log main;
    access_log      /foo/bar/new.log main;
    """
    fs_unix.map_file_fh("etc/nginx/nginx.conf", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("foo/bar/new.log", BytesIO(b"New"))
    fs_unix.map_file_fh("foo/bar/old.log", BytesIO(b"Old"))
    target_unix.add_plugin(NginxPlugin)

    log_paths = target_unix.nginx.get_log_paths()
    assert str(log_paths[0]) == "/foo/bar/old.log"
    assert str(log_paths[1]) == "/foo/bar/new.log"



import re

from dissect.target.plugins.os.windows.defender_helpers.defender_records import (
    DefenderMPLogBMTelemetryRecord,
    DefenderMPLogDetectionAddRecord,
    DefenderMPLogDetectionEventRecord,
    DefenderMPLogEMSRecord,
    DefenderMPLogExclusionRecord,
    DefenderMPLogLowfiRecord,
    DefenderMPLogMinFilBlockedFileRecord,
    DefenderMPLogMinFilUSSRecord,
    DefenderMPLogOriginalFileNameRecord,
    DefenderMPLogProcessImageRecord,
    DefenderMPLogResourceScanRecord,
    DefenderMPLogRTPRecord,
    DefenderMPLogThreatActionRecord,
    DefenderMPLogThreatRecord,
)

DEFENDER_MPLOG_TS_PATTERN = r"(?P<ts>[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{3}Z) "

# Loosely based on https://github.com/Intrinsec/mplog_parser but feel free to add patterns

DEFENDER_MPLOG_PATTERNS = [
    # Process Image
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r"ProcessImageName: (?P<process_image_name>.*), ",
                    r"Pid: (?P<pid>\d*), ",
                    r"TotalTime: (?P<total_time>\d*), ",
                    r"Count: (?P<count>\d*), ",
                    r"MaxTime: (?P<max_time>\d*), ",
                    r"MaxTimeFile: (?P<max_time_file>.*), ",
                    r"EstimatedImpact: (?P<estimated_impact>\d*)",
                ]
            )
        ),
        DefenderMPLogProcessImageRecord,
    ),
    # Mini-filter Unsuccessful scan status
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r"\[Mini-filter\] (Unsuccessful scan status)[^:]*: (?P<path>.+) ",
                    r"Process: (?P<process>.+), ",
                    r"Status: (?P<status>.+), ",
                    r"State: (?P<state>.+), ",
                    r"ScanRequest (?P<scan_request>.+), ",
                    r"FileId: (?P<file_id>.+), ",
                    r"Reason: (?P<reason>.+), ",
                    r"IoStatusBlockForNewFile: (?P<io_status_block_for_new_file>.+), ",
                    r"DesiredAccess:(?P<desired_access>.+), ",
                    r"FileAttributes:(?P<file_attributes>.+), ",
                    r"ScanAttributes:(?P<scan_attributes>.+), ",
                    r"AccessStateFlags:(?P<access_state_flags>.+), ",
                    r"BackingFileInfo: (?P<backing_file_info>.+)",
                ]
            )
        ),
        DefenderMPLogMinFilUSSRecord,
    ),
    # EMS Scan
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r".*",
                    r"process: (?P<process>\w*) ",
                    r"pid: (?P<pid>\d*), ",
                    r"sigseq: (?P<sigseq>\w*), ",
                    r"sendMemoryScanReport: (?P<send_memory_scan_report>\d*), ",
                    r"source: (?P<source>\d*)",
                ]
            )
        ),
        DefenderMPLogEMSRecord,
    ),
    # Original filename
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r".*",
                    r"original file name \"(?P<original_file_name>.*)\" ",
                    r"for \"(?P<full_path>.*)\", ",
                    r"hr=(?P<hr>\w*)",
                ]
            )
        ),
        DefenderMPLogOriginalFileNameRecord,
    ),
    # Mini-filter Blocked file
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r".*",
                    r"\[Mini-filter\] Blocked file: (?P<blocked_file>.+) ",
                    r"Process: (?P<process>.+), ",
                    r"Status: (?P<status>.+), ",
                    r"State: (?P<state>.+), ",
                    r"ScanRequest (?P<scan_request>.+), ",
                    r"FileId: (?P<file_id>.+), ",
                    r"Reason: (?P<reason>.+), ",
                    r"IoStatusBlockForNewFile: (?P<io_status_block_for_new_file>.+), ",
                    r"DesiredAccess:(?P<desired_access>.+), ",
                    r"FileAttributes:(?P<file_attributes>.+), ",
                    r"ScanAttributes:(?P<scan_attributes>.+), ",
                    r"AccessStateFlags:(?P<access_state_flags>.+), ",
                    r"BackingFileInfo: (?P<backing_file_info>.+)",
                ]
            )
        ),
        DefenderMPLogMinFilBlockedFileRecord,
    ),
    # Exclusion
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r"\[Exclusion\] (?P<full_path_with_drive_letter>.+) ",
                    r"-> (?P<full_path_with_device_path>.+)",
                ]
            )
        ),
        DefenderMPLogExclusionRecord,
    ),
    # Lowfi
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r".*",
                    r"lowfi: (?P<lowfi>.+)",
                ]
            )
        ),
        DefenderMPLogLowfiRecord,
    ),
    # Detection add
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r".*",
                    r"DETECTION_ADD\S* (?P<detection>.*)",
                ]
            )
        ),
        DefenderMPLogDetectionAddRecord,
    ),
    # Threat
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r".*",
                    r"threat: (?P<threat>.*)",
                ]
            )
        ),
        DefenderMPLogThreatRecord,
    ),
    # Detection event
    (
        re.compile(
            "".join(
                [
                    DEFENDER_MPLOG_TS_PATTERN,
                    r".*",
                    r"DETECTIONEVENT MPSOURCE_\S+ HackTool:(?P<threat_type>.*) file:(?P<command>.*)",
                ]
            )
        ),
        DefenderMPLogDetectionEventRecord,
    ),
]


DEFENDER_MPLOG_BLOCK_PATTERNS = [
    (
        re.compile(r"Begin Resource Scan"),
        re.compile(r"End Scan"),
        re.compile(
            "".join(
                [
                    r"Begin Resource Scan.*\n",
                    r"Scan ID:(?P<scan_id>[^\n]+)\n",
                    r"Scan Source:(?P<scan_source>\d+)\n",
                    r"Start Time:(?P<start_time>[0-9\-\:\s]*)\n",
                    r"End Time:(?P<end_time>[0-9\-\:\s]*)\n",
                    r".*",
                    r"Resource Schema:(?P<resource_schema>[^\n]+)\n",
                    r"Resource Path:(?P<resource_path>[^\n]+)\n",
                    r"Result Count:(?P<result_count>\d+)\n",
                    r"(?P<rest>.*)\n",
                    r"End Scan",
                ]
            ),
            re.MULTILINE | re.DOTALL,
        ),
        DefenderMPLogResourceScanRecord,
    ),
    # Threat actions
    (
        re.compile(r"Beginning threat actions"),
        re.compile(r"Finished threat actions"),
        re.compile(
            "".join(
                [
                    r"Beginning threat actions\n",
                    r"Start time:(?P<ts>[0-9\-\:\s]*)\n",
                    r"(?P<rest>.*)\n",
                    r"Finished threat actions",
                ]
            ),
            re.MULTILINE | re.DOTALL,
        ),
        DefenderMPLogThreatActionRecord,
    ),
    # RTP
    (
        re.compile(r"\*\*RTP Perf Log\*\*"),
        re.compile(r"\*\*END RTP Perf Log\*\*"),
        re.compile(
            "".join(
                [
                    r"\*+RTP Perf Log\*+\n",
                    r"RTP Start:(?P<ts>.*)\n",
                    r"Last Perf:(?P<last_perf>.*)\n",
                    r"First RTP Scan:(?P<first_rtp_scan>.*)\n",
                    r"Plugin States:(?P<plugin_states>.*)\n",
                    r"Process Exclusions:\n(?P<process_exclusions>.*)",
                    r"Path Exclusions:\n(?P<path_exclusions>.*)",
                    r"Ext Exclusions:\n(?P<ext_exclusions>.*)",
                    r"Worker Threads",
                ]
            ),
            re.MULTILINE | re.DOTALL,
        ),
        DefenderMPLogRTPRecord,
    ),
    # BM Telemetry (block)
    (
        re.compile(r"BEGIN BM telemetry"),
        re.compile(r"END BM telemetry"),
        re.compile(
            "".join(
                [
                    r"BEGIN BM telemetry\n",
                    r"(GUID):(?P<guid>.+)\n",
                    r"(SignatureID):(?P<signature_id>.+)\n",
                    r"(SigSha):(?P<sigsha>.+)\n",
                    r"(ThreatLevel):(?P<threat_level>.+)\n",
                    r"(ProcessID):(?P<process_id>.+)\n",
                    r"(ProcessCreationTime):(?P<process_creation_time>.+)\n",
                    r"(SessionID):(?P<session_id>.+)\n",
                    r"(CreationTime):(?P<ts>.+)\n",
                    r"(ImagePath):(?P<image_path>.+)\n",
                    r"(Taint Info):(?P<taint_info>.+)\n",
                    r"(Operations):(?P<operations>.+)\n",
                    r"END BM telemetry",
                ]
            )
        ),
        DefenderMPLogBMTelemetryRecord,
    ),
]

DEFENDER_MPLOG_LINE = re.compile(r"^\s+(.*)$", re.MULTILINE)



from __future__ import annotations

import io
import textwrap
from logging import getLogger
from typing import Any, BinaryIO, Iterator, Optional

from dissect.target import Target
from dissect.target.exceptions import ConfigurationParsingError, FileNotFoundError
from dissect.target.filesystem import FilesystemEntry, VirtualFilesystem
from dissect.target.helpers import fsutil
from dissect.target.helpers.configutil import ConfigurationParser, parse

log = getLogger(__name__)


class ConfigurationFilesystem(VirtualFilesystem):
    """A special :class:`.Filesystem` class that allows you to browse and interact with configuration files
    as ``directories`` and ``files`` by parsing the file into key/value pairs.

    Depending on the ``value`` of a configuration file's ``key``, it will act like a ``directory`` or ``file``:
        * When the ``key`` contains sub-values (dictionary), it will act like a ``directory``.
        * Otherwise it will act like a ``file``.

    Examples:
        >>> fs = ConfigurationFilesystem(target, "/etc")
        >>> entry = fs.get("xattr.conf")
        <ConfigurationEntry path=/etc/xattr.conf value=<dissect.target.helpers.configutil.Default object at 0x115683280>
        >>> entry.listdir() # listed entries are the keys in the configuration file
        [...
        'system.posix_acl_access',
        'system.posix_acl_default',
        'trusted.SGI_ACL_DEFAULT',
        'trusted.SGI_ACL_FILE',
        ...]
        >>> entry.get("system.posix_acl_access") # returns the value of the key
        <ConfigurationEntry path=/etc/xattr.conf/system.posix_acl_access value=permissions>
        >>> entry.get("system.posix_acl_access").open().read() # returns the raw value of the key
        b'permissions\\n'

    """

    __type__: str = "META:configuration"

    def __init__(self, target: Target, path: str, **kwargs):
        super().__init__(**kwargs)
        self.root.top = target.fs.get(path)

    def _get_till_file(self, path: str, relentry: FilesystemEntry | None) -> tuple[list[str], FilesystemEntry]:
        """Searches for the file entry that is pointed to by ``path``.

        The ``path`` could contain ``key`` entries too, so it searches for the entry from
        the start of the path.

        Returns:
            A list of ``parts`` containing keys: [keys, into, the, file].
            And the resolved entry: Entry(filename)
        """

        entry = relentry or self.root
        root_path = relentry.path if relentry else self.root.top.path

        # Calculate the relative path
        relpath = fsutil.relpath(path, root_path, alt_separator=self.alt_separator)
        path = fsutil.normalize(relpath, alt_separator=self.alt_separator).strip("/")

        if not path:
            return [], entry

        parts = path.split("/")

        for idx, part in enumerate(parts, start=1):
            # Resolve link
            if entry.is_symlink():
                entry = entry.readlink_ext()

            if part == ".":
                continue
            elif part == "..":
                entry = entry.up or self.root
                continue

            if entry is self.root:
                entry = self.root.top

            entry = entry.get(part)
            if entry.is_file():
                break

        return parts[idx:], entry

    def get(self, path: str, relentry: Optional[FilesystemEntry] = None, *args, **kwargs) -> ConfigurationEntry:
        """Retrieve a :class:`ConfigurationEntry` relative to the root or ``relentry``.

        Raises:
            FileNotFoundError: if it could not find the entry.
        """
        parts, entry = self._get_till_file(path, relentry)

        if entry.is_dir():
            return ConfigurationEntry(self, entry.path, entry, None)

        entry = self._convert_entry(entry, *args, **kwargs)

        for part in parts:
            _prev_entry = entry
            entry = entry.get(part)
            if entry is None:
                raise FileNotFoundError(f"{part!r} not found in {_prev_entry}.")

        return entry

    def _convert_entry(self, file_entry: FilesystemEntry, *args, **kwargs) -> ConfigurationEntry:
        """Creates a :class:`ConfigurationEntry` from a ``file_entry``.

        If an error occurs during the parsing of the file contents,
        the original ``file_entry`` is returned.
        """
        entry = file_entry
        config_parser = None
        try:
            config_parser = parse(entry, *args, **kwargs)
        except ConfigurationParsingError as e:
            # If a parsing error gets created, it should return the `entry`
            log.debug("Error when parsing %s with message '%s'", entry.path, e)

        return ConfigurationEntry(self, entry.path, entry, config_parser)


class ConfigurationEntry(FilesystemEntry):
    """A Special filesystem entry.

    Behaves like a ``directory`` when :attr:`parser_items` is a :class:`.ConfigurationParser` or a ``dict``.
    Behaves like a ``file`` otherwise.

    Args:
        parser_items: A dict-like object containing all configuration entries and values.
            In most cases this is either a :class:`.ConfigurationParser` or ``dict``.
            Otherwise, its the entry's value

    Examples:
        >>> fs = ConfigurationFilesystem(target, "/etc")
        >>> entry = fs.get("xattr.conf")
        <ConfigurationEntry path=/etc/xattr.conf value=<dissect.target.helpers.configutil.Default object at 0x115683280>
        >>> entry.listdir() # listed entries are the keys in the configuration file
        [...
        'system.posix_acl_access',
        'system.posix_acl_default',
        'trusted.SGI_ACL_DEFAULT',
        'trusted.SGI_ACL_FILE',
        ...]
        >>> entry.as_dict()
        {
            ...
            "system.posix_acl_access" : "...",
            ...
        }
        >>> entry.get("system.posix_acl_access") # returns the value of the key
        <ConfigurationEntry path=/etc/xattr.conf/system.posix_acl_access value=permissions>
        >>> entry.get("system.posix_acl_access").open().read() # returns the raw value of the key
        b'permissions\\n'
    """

    def __init__(
        self,
        fs: ConfigurationFilesystem,
        path: str,
        entry: FilesystemEntry,
        parser_items: dict | ConfigurationParser | str | list | None = None,
    ) -> None:
        super().__init__(fs, path, entry)
        self.parser_items = parser_items

    def __getitem__(self, item: str) -> ConfigurationEntry:
        return self.parser_items[item]

    def __repr__(self) -> str:
        output = f"path={self.path}"

        if not isinstance(self.parser_items, dict):
            output += f" value={self.parser_items}"

        return f"<{self.__class__.__name__} {output}"

    def get(self, key, default: Any | None = None) -> ConfigurationEntry | Any | None:
        """Gets the dictionary key that belongs to this entry using ``key``.
        Behaves like ``dictionary.get()``.

        Args:
            ``key``: A dictionary key that is inside :attr:`parser_items`.
            ``default``: The default value to return if ``key`` is not inside this entry.

        Returns:
            a :class:`ConfigurationEntry` when ``key`` is present, otherwise its ``default``.
        """
        # Check for path in config entry
        if not key:
            raise TypeError("key should be defined")

        path = fsutil.join(self.path, key, alt_separator=self.fs.alt_separator)

        if self.parser_items and key in self.parser_items:
            return ConfigurationEntry(
                self.fs,
                path,
                self.entry,
                self.parser_items[key],
            )

        if self.entry.is_dir():
            return self.fs.get(path, self.entry)

        return default

    def _write_value_mapping(self, values: dict[str, Any], indentation_nr: int = 0) -> str:
        """Internal function to transform the ``values`` dictionary to a string representation.

        Args:
            values: A dictionary or its contents.
            indentation_nr: How much indentation this entry should receive.

        Returns:
            An indented string containing all the information inside ``values``.
        """
        prefix = " " * indentation_nr
        output_buffer = io.StringIO()

        if isinstance(values, list):
            # Explicitly convert the list values to strings
            _text = "\n".join(str(val) for val in values)
            output_buffer.write(textwrap.indent(text=_text, prefix=prefix))
        elif hasattr(values, "keys"):
            for key, value in values.items():
                output_buffer.write(textwrap.indent(key, prefix=prefix) + "\n")
                output_buffer.write(self._write_value_mapping(value, indentation_nr + 4))
        else:
            output_buffer.write(textwrap.indent(str(values), prefix=prefix) + "\n")

        return output_buffer.getvalue()

    def open(self) -> BinaryIO:
        """Open this :class:`ConfigurationEntry`.

        If :attr:`parser_items` is a :class:`.ConfigurationParser`,
        it will ``open`` the underlying entry.

        Returns:
            A file-like object holding a byte representation of :attr:`parser_items`.
        """

        if isinstance(self.parser_items, ConfigurationParser):
            # Currently trying to open the underlying entry
            return self.entry.open()

        if isinstance(self.parser_items, bytes):
            return io.BytesIO(self.parser_items)

        output_data = self._write_value_mapping(self.parser_items)
        return io.BytesIO(bytes(output_data, "utf-8"))

    def iterdir(self) -> Iterator[str]:
        for entry in self.scandir():
            yield entry.name

    def scandir(self) -> Iterator[ConfigurationEntry]:
        """Return the items inside :attr:`parser_items` as ``ConfigurationEntries``."""
        if self.is_file():
            raise NotADirectoryError()

        if self.parser_items is None and self.entry.is_dir():
            for entry in self.entry.scandir():
                yield ConfigurationEntry(self.fs, entry.name, entry, None)
            return

        for key, values in self.parser_items.items():
            yield ConfigurationEntry(self.fs, key, self.entry, values)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        return not self.is_dir(follow_symlinks)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Returns whether this :class:`ConfigurationEntry` can be considered a directory."""
        # if self.parser_items has keys (thus sub-values), we can consider it a directory.
        return (self.parser_items is None and self.entry.is_dir()) or hasattr(self.parser_items, "keys")

    def is_symlink(self) -> bool:
        """Return whether this :class:`ConfigurationEntry` is a symlink or not.

        Returns:
            False, as ``ConfigurationEntries`` are never symlinks.
        """
        # ConfigurationEntries are already resolved, so are never symlinks.
        return False

    def exists(self, key: str) -> bool:
        """Return whether the underlying :class:`.FilesystemEntry` :attr:`entry` and
        supplied ``key`` exists inside this :class:`ConfigurationEntry`.

        Returns:
            Whether the ``entry`` and ``key`` exists
        """
        return self.parser_items and self.entry.exists() and key in self.parser_items

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Returns the stat from the underlying :class:`.FilesystemEntry` :attr:`entry`."""
        return self.entry.stat(follow_symlinks)

    def lstat(self) -> fsutil.stat_result:
        """Returns the lstat from the underlying :class:`.FilesystemEntry` :attr:`entry`."""

        return self.entry.lstat()

    def as_dict(self) -> dict:
        """Returns :attr:`parser_items` as a dictionary."""
        if isinstance(self.parser_items, ConfigurationParser):
            return self.parser_items.parsed_data
        return self.parser_items



from typing import Iterator

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.dpapi.keyprovider.keyprovider import (
    KeyProviderPlugin,
)

defaultpassword_def = """
struct DefaultPassword {
    DWORD   length;
    char    flags[4*3];
    WCHAR   data[length/2];
    char    checksum_or_guid[0x10];
};
"""

c_defaultpassword = cstruct().load(defaultpassword_def)


class LSADefaultPasswordKeyProviderPlugin(KeyProviderPlugin):
    """Windows LSA DefaultPassword key provider plugin."""

    __namespace__ = "_dpapi_keyprovider_lsa_defaultpassword"

    def check_compatible(self) -> None:
        if not self.target.has_function("lsa"):
            raise UnsupportedPluginError("LSA plugin not available on target")

    @export(output="yield")
    def keys(self) -> Iterator[tuple[str, str]]:
        """Yield Windows LSA DefaultPassword strings."""
        if default_pass := self.target.lsa._secrets.get("DefaultPassword"):
            try:
                value = c_defaultpassword.DefaultPassword(default_pass).data
            except Exception:
                self.target.log.warning("Failed to parse LSA DefaultPassword value")
                return
            yield self.__namespace__, value



from __future__ import annotations

from dissect.target.filesystem import Filesystem
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin
from dissect.target.target import Target


class FreeBsdPlugin(BsdPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self._os_release = self._parse_os_release("/bin/freebsd-version*")

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if fs.exists("/net") and (fs.exists("/.sujournal") or fs.exists("/entropy")):
                return fs

        return None

    @export(property=True)
    def version(self) -> str | None:
        return self._os_release.get("USERLAND_VERSION")



from __future__ import annotations

import re
from dataclasses import dataclass, field
from datetime import datetime, timezone
from ipaddress import ip_address, ip_interface
from typing import TYPE_CHECKING, Any, Iterator, Literal, NamedTuple

from dissect.target.helpers import configutil
from dissect.target.helpers.record import UnixInterfaceRecord
from dissect.target.helpers.utils import to_list
from dissect.target.plugins.general.network import NetworkPlugin

if TYPE_CHECKING:
    from ipaddress import IPv4Address, IPv4Interface, IPv6Address, IPv6Interface

    from dissect.target import Target
    from dissect.target.target import TargetPath

    NetAddress = IPv4Address | IPv6Address
    NetInterface = IPv4Interface | IPv6Interface


class LinuxNetworkPlugin(NetworkPlugin):
    """Linux network interface plugin."""

    def _interfaces(self) -> Iterator[UnixInterfaceRecord]:
        """Try all available network configuration managers and aggregate the results."""
        for manager_cls in MANAGERS:
            manager: LinuxNetworkConfigParser = manager_cls(self.target)
            yield from manager.interfaces()


VlanIdByInterface = dict[str, set[int]]


class LinuxNetworkConfigParser:
    def __init__(self, target: Target):
        self._target = target

    def _config_files(self, config_paths: list[str], glob: str) -> list[TargetPath]:
        """Returns all configuration files in config_paths matching the given extension."""
        all_files = []
        for config_path in config_paths:
            paths = self._target.fs.path(config_path).glob(glob)
            all_files.extend(config_file for config_file in paths if config_file.is_file())

        return sorted(all_files, key=lambda p: p.stem)

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        """Parse network interfaces from configuration files."""
        yield from ()


class NetworkManagerConfigParser(LinuxNetworkConfigParser):
    """NetworkManager configuration parser.

    NetworkManager configuration files are generally in an INI-like format.
    Note that Red Hat and Fedora deprecated ifcfg files.
    Documentation: https://networkmanager.dev/docs/api/latest/nm-settings-keyfile.html
    """

    config_paths: list[str] = [
        "/etc/NetworkManager/system-connections/",
        "/usr/lib/NetworkManager/system-connections/",
        "/run/NetworkManager/system-connections/",
    ]

    @dataclass
    class ParserContext:
        source: str
        uuid: str | None = None
        last_connected: datetime | None = None
        name: str | None = None
        mac_address: str | None = None
        type: str = ""
        dns: set[NetAddress] = field(default_factory=set)
        ip_interfaces: set[NetInterface] = field(default_factory=set)
        gateways: set[NetAddress] = field(default_factory=set)
        dhcp_ipv4: bool = False
        dhcp_ipv6: bool = False
        vlan: set[int] = field(default_factory=set)

        def to_record(self) -> UnixInterfaceRecord:
            return UnixInterfaceRecord(
                source=self.source,
                last_connected=self.last_connected,
                name=self.name,
                mac=[self.mac_address] if self.mac_address else [],
                type=self.type,
                dhcp_ipv4=self.dhcp_ipv4,
                dhcp_ipv6=self.dhcp_ipv6,
                dns=list(self.dns),
                ip=[interface.ip for interface in self.ip_interfaces],
                network=[interface.network for interface in self.ip_interfaces],
                gateway=list(self.gateways),
                vlan=list(self.vlan),
                configurator="NetworkManager",
            )

    def interfaces(self) -> Iterator[UnixInterfaceRecord]:
        connections: list[NetworkManagerConfigParser.ParserContext] = []
        vlan_id_by_interface: VlanIdByInterface = {}

        for connection_file_path in self._config_files(self.config_paths, "*"):
            try:
                config = configutil.parse(connection_file_path, hint="ini")
                context = self.ParserContext(source=connection_file_path)
                common_section: dict[str, str] = config.get("connection", {})
                context.type = common_section.get("type", "")
                sub_type: dict[str, str] = config.get(context.type, {})

                if context.type == "vlan":
                    self._parse_vlan(sub_type, vlan_id_by_interface)
                    continue

                for ip_version in ["ipv4", "ipv6"]:
                    ip_section: dict[str, str] = config.get(ip_version, {})
                    for key, value in ip_section.items():
                        self._parse_ip_section_key(key, value, context, ip_version)

                context.name = common_section.get("interface-name")
                context.mac_address = sub_type.get("mac-address")
                context.uuid = common_section.get("uuid")
                context.source = str(connection_file_path)
                context.last_connected = self._parse_lastconnected(common_section.get("timestamp", ""))

                connections.append(context)

            except Exception as e:
                self._target.log.warning("Error parsing network config file %s", connection_file_path)
                self._target.log.debug("", exc_info=e)

        for connection in connections:
            vlan_ids_from_interface = vlan_id_by_interface.get(connection.name, set())
            connection.vlan.update(vlan_ids_from_interface)

            vlan_ids_from_uuid = vlan_id_by_interface.get(connection.uuid, set())
            connection.vlan.update(vlan_ids_from_uuid)

            yield connection.to_record()

    def _parse_route(self, route: str) -> NetAddress | None:
        """Parse a route and return gateway IP address."""
        if (elements := route.split(",")) and len(elements) > 1:
            return ip_address(elements[1])

        return None

    def _parse_lastconnected(self, last_connected: str) -> datetime | None:
        """Parse last connected timestamp."""
        if not last_connected:
            return None

        return datetime.fromtimestamp(int(last_connected), timezone.utc)

    def _parse_ip_section_key(
        self, key: str, value: str, context: ParserContext, ip_version: Literal["ipv4", "ipv6"]
    ) -> None:
        if not (trimmed := value.strip()):
            return

        if key == "dns":
            context.dns.update(ip_address(addr) for addr in trimmed.split(";") if addr)
        elif key.startswith("address"):
            # Undocumented: single gateway on address line. Observed when running:
            # nmcli connection add type ethernet ... ip4 192.168.2.138/24 gw4 192.168.2.1
            ip, *gateway = trimmed.split(",", 1)
            context.ip_interfaces.add(ip_interface(ip))
            if gateway:
                context.gateways.add(ip_address(gateway[0]))
        elif key.startswith("gateway"):
            context.gateways.add(ip_address(trimmed))
        elif key == "method":
            if ip_version == "ipv4":
                context.dhcp_ipv4 = trimmed == "auto"
            elif ip_version == "ipv6":
                context.dhcp_ipv6 = trimmed == "auto"
        elif key.startswith("route"):
            if gateway := self._parse_route(value):
                context.gateways.add(gateway)

    def _parse_vlan(self, sub_type: dict[str, Any], vlan_id_by_interface: VlanIdByInterface) -> None:
        parent_interface = sub_type.get("parent")
        vlan_id = sub_type.get("id")
        if not parent_interface or not vlan_id:
            return

        ids = vlan_id_by_interface.setdefault(parent_interface, set())
        ids.add(int(vlan_id))


class SystemdNetworkConfigParser(LinuxNetworkConfigParser):
    """Systemd network configuration parser.

    Systemd network configuration files are generally in an INI-like format with some quirks.
    Note that drop-in directories are not yet supported.

    Documentation: https://www.freedesktop.org/software/systemd/man/latest/systemd.network.html
    """

    config_paths: list[str] = [
        "/etc/systemd/network/",
        "/run/systemd/network/",
        "/usr/lib/systemd/network/",
        "/usr/local/lib/systemd/network/",
    ]

    class DhcpConfig(NamedTuple):
        ipv4: bool
        ipv6: bool

    # Can be enclosed in brackets for IPv6. Can also have port, iface name, and SNI, which we ignore.
    # Example: [1111:2222::3333]:9953%ifname#example.com
    dns_ip_patttern = re.compile(
        r"(?P<withoutBrackets>(?:\d{1,3}\.){3}\d{1,3})|\[(?P<withBrackets>\[?[0-9a-fA-F:]+\]?)\]"
    )

    def interfaces(self) -> Iterator:
        virtual_networks = self._parse_virtual_networks()
        yield from self._parse_networks(virtual_networks)

    def _parse_virtual_networks(self) -> VlanIdByInterface:
        """Parse virtual network configurations from systemd network configuration files."""

        virtual_networks: VlanIdByInterface = {}
        for config_file in self._config_files(self.config_paths, "*.netdev"):
            try:
                virtual_network_config = configutil.parse(config_file, hint="systemd")
                net_dev_section: dict[str, str] = virtual_network_config.get("NetDev", {})
                if net_dev_section.get("Kind") != "vlan":
                    continue

                vlan_id = virtual_network_config.get("VLAN", {}).get("Id")
                if (name := net_dev_section.get("Name")) and vlan_id:
                    vlan_ids = virtual_networks.setdefault(name, set())
                    vlan_ids.add(int(vlan_id))
            except Exception as e:
                self._target.log.warning("Error parsing virtual network config file %s", config_file)
                self._target.log.debug("", exc_info=e)

        return virtual_networks

    def _parse_networks(self, virtual_networks: VlanIdByInterface) -> Iterator[UnixInterfaceRecord]:
        """Parse network configurations from systemd network configuration files."""
        for config_file in self._config_files(self.config_paths, "*.network"):
            try:
                config = configutil.parse(config_file, hint="systemd")

                match_section: dict[str, str] = config.get("Match", {})
                network_section: dict[str, str] = config.get("Network", {})
                link_section: dict[str, str] = config.get("Link", {})

                ip_interfaces: set[NetInterface] = set()
                gateways: set[NetAddress] = set()
                dns: set[NetAddress] = set()
                mac_addresses: set[str] = set()

                if link_mac := link_section.get("MACAddress"):
                    mac_addresses.add(link_mac)
                mac_addresses.update(match_section.get("MACAddress", "").split())
                mac_addresses.update(match_section.get("PermanentMACAddress", "").split())

                dns_value = to_list(network_section.get("DNS", []))
                dns.update(self._parse_dns_ip(dns_ip) for dns_ip in dns_value)

                address_value = to_list(network_section.get("Address", []))
                ip_interfaces.update(ip_interface(addr) for addr in address_value)

                gateway_value = to_list(network_section.get("Gateway", []))
                gateways.update(ip_address(gateway) for gateway in gateway_value)

                vlan_ids: set[int] = set()
                vlan_names = to_list(network_section.get("VLAN", []))
                for vlan_name in vlan_names:
                    if ids := virtual_networks.get(vlan_name):
                        vlan_ids.update(ids)

                # There are possibly multiple route sections, but they are collapsed into one by the parser.
                route_section: dict[str, Any] = config.get("Route", {})
                gateway_values = to_list(route_section.get("Gateway", []))
                gateways.update(filter(None, map(self._parse_gateway, gateway_values)))

                dhcp_ipv4, dhcp_ipv6 = self._parse_dhcp(network_section.get("DHCP"))

                yield UnixInterfaceRecord(
                    source=str(config_file),
                    type=match_section.get("Type"),
                    enabled=None,  # Unknown, dependent on run-time state
                    dhcp_ipv4=dhcp_ipv4,
                    dhcp_ipv6=dhcp_ipv6,
                    name=match_section.get("Name"),
                    dns=list(dns),
                    mac=list(mac_addresses),
                    ip=[interface.ip for interface in ip_interfaces],
                    network=[interface.network for interface in ip_interfaces],
                    gateway=list(gateways),
                    vlan=list(vlan_ids),
                    configurator="systemd-networkd",
                )
            except Exception as e:
                self._target.log.warning("Error parsing network config file %s", config_file)
                self._target.log.debug("", exc_info=e)

    def _parse_dns_ip(self, address: str) -> NetAddress:
        """Parse DNS address from systemd network configuration file.

        The optional brackets and port number make this hard to parse.
        See https://www.freedesktop.org/software/systemd/man/latest/systemd.network.html and search for DNS.
        """

        if match := self.dns_ip_patttern.search(address):
            return ip_address(match.group("withoutBrackets") or match.group("withBrackets"))

        raise ValueError(f"Invalid DNS address format: {address}")

    def _parse_dhcp(self, value: str | None) -> DhcpConfig:
        """Parse DHCP value from systemd network configuration file to a named tuple (ipv4, ipv6)."""

        if value is None or value == "no":
            return self.DhcpConfig(ipv4=False, ipv6=False)
        elif value == "yes":
            return self.DhcpConfig(ipv4=True, ipv6=True)
        elif value == "ipv4":
            return self.DhcpConfig(ipv4=True, ipv6=False)
        elif value == "ipv6":
            return self.DhcpConfig(ipv4=False, ipv6=True)

        raise ValueError(f"Invalid DHCP value: {value}")

    def _parse_gateway(self, value: str | None) -> NetAddress | None:
        if (not value) or (value in {"_dhcp4", "_ipv6ra"}):
            return None

        return ip_address(value)


MANAGERS = [NetworkManagerConfigParser, SystemdNetworkConfigParser]



from dissect.target.plugin import InternalNamespacePlugin


class KeyProviderPlugin(InternalNamespacePlugin):
    __namespace__ = "_dpapi_keyprovider"

    def check_compatible(self) -> None:
        return None



from pathlib import Path
from unittest.mock import MagicMock, call, patch

from dissect.target.loaders.ovf import OvfLoader
from dissect.target.target import Target


@patch("dissect.target.loaders.ovf.container")
@patch("dissect.target.loaders.ovf.ovf.OVF")
def test_ovf_loader(OVF: MagicMock, container: MagicMock, target_bare: Target, tmp_path: Path):
    (tmp_path / "test.ovf").touch()

    OVF.return_value = OVF
    OVF.disks.return_value = ["disk.vmdk"]
    container.open.return_value = MagicMock()

    ovf_loader = OvfLoader(tmp_path / "test.ovf")
    ovf_loader.map(target_bare)

    assert len(target_bare.disks) == 1
    assert container.open.mock_calls == [call(tmp_path.resolve() / "disk.vmdk")]



import socket
import ssl
from struct import unpack
from unittest.mock import MagicMock, call, patch

from dissect.target.loaders.remote import RemoteStream, RemoteStreamConnection


@patch.object(ssl, "SSLContext", autospec=True)
@patch.object(socket, "socket", autospec=True)
def test_remote_loader_stream(mock_socket_class: MagicMock, mock_context: MagicMock) -> None:
    rsc = RemoteStreamConnection("remote://127.0.0.1", 9001, options={"ca": "A", "key": "B", "crt": "C"})
    assert rsc.is_connected() is False
    rsc.connect()

    rsc.remote_disk_data = b"ABC"
    rsc.remote_disk_response = None

    def send(data: bytes) -> None:
        _, offset, read = unpack(">BQQ", data)
        rsc.remote_disk_response = rsc.remote_disk_data[offset : offset + read]
        return len(data)

    def receive(num: bytes) -> bytes:
        return rsc.remote_disk_response

    rsc._ssl_sock.send = send
    rsc._ssl_sock.recv = receive
    rs = RemoteStream(rsc, 0, 3)
    rs.align = 1
    rs.seek(1)
    data = rs.read(2)
    rs.close()
    mock_socket_class.assert_called_with(socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_STREAM)
    expected = [
        call(ssl.PROTOCOL_TLSv1_2),
        call().load_default_certs(),
        call().load_cert_chain(certfile="C", keyfile="B"),
        call().load_verify_locations("A"),
        call().wrap_socket(rsc._socket, server_hostname="remote://127.0.0.1"),
        call().wrap_socket().connect(("remote://127.0.0.1", 9001)),
    ]
    assert data == b"BC"
    assert mock_context.mock_calls == expected
    assert rs.tell() == 3
    assert rsc.is_connected() is True


@patch.object(ssl, "SSLContext", autospec=False)
@patch.object(socket, "socket", autospec=True)
def test_remote_loader_stream_embedded(mock_socket_class: MagicMock, mock_context: MagicMock) -> None:
    RemoteStreamConnection.configure("K", "C")
    RemoteStreamConnection("remote://127.0.0.1", 9001)
    mock_context.assert_has_calls([call().load_cert_chain_str(certfile="C", keyfile="K")])



import io
from pathlib import Path
from typing import BinaryIO, Union

from dissect.hypervisor import vhd

from dissect.target.container import Container


class VhdContainer(Container):
    __type__ = "vhd"

    def __init__(self, fh: Union[BinaryIO, Path], *args, **kwargs):
        f = fh
        if not hasattr(fh, "read"):
            f = fh.open("rb")
        self.vhd = vhd.VHD(f)

        super().__init__(fh, self.vhd.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: Union[list, BinaryIO]) -> bool:
        try:
            fh.seek(-512, io.SEEK_END)
            return b"conectix" in fh.read(9)
        except OSError:
            return False

    @staticmethod
    def detect_path(path: Path, original: Union[list, BinaryIO]) -> bool:
        return path.suffix.lower() == ".vhd"

    def read(self, length: int) -> bytes:
        return self.vhd.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.vhd.seek(offset, whence)

    def tell(self) -> int:
        return self.vhd.tell()

    def close(self) -> None:
        pass



import os
from functools import reduce
from pathlib import Path
from typing import Iterator, Optional
from unittest.mock import Mock, patch

import pytest
from docutils.core import publish_string
from docutils.utils import SystemMessage
from flow.record import Record

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import EmptyRecord, create_extended_descriptor
from dissect.target.plugin import (
    PLUGINS,
    InternalNamespacePlugin,
    InternalPlugin,
    NamespacePlugin,
    OSPlugin,
    Plugin,
    PluginFunction,
    alias,
    environment_variable_paths,
    export,
    find_plugin_functions,
    get_external_module_paths,
    plugins,
    save_plugin_import_failure,
)
from dissect.target.plugins.general.default import DefaultPlugin
from dissect.target.target import Target


def test_save_plugin_import_failure():
    test_trace = ["test-trace"]
    test_module_name = "test-module"

    with patch("traceback.format_exception", Mock(return_value=test_trace)):
        with patch("dissect.target.plugin.PLUGINS", new_callable=dict) as MOCK_PLUGINS:
            MOCK_PLUGINS["_failed"] = []
            save_plugin_import_failure(test_module_name)

            assert len(MOCK_PLUGINS["_failed"]) == 1
            assert MOCK_PLUGINS["_failed"][0].get("module") == test_module_name
            assert MOCK_PLUGINS["_failed"][0].get("stacktrace") == test_trace


@pytest.mark.parametrize(
    "env_value, expected_output",
    [
        (None, []),
        ("", []),
        (":", [Path(""), Path("")]),
    ],
)
def test_load_environment_variable(env_value, expected_output):
    with patch.object(os, "environ", {"DISSECT_PLUGINS": env_value}):
        assert environment_variable_paths() == expected_output


def test_load_module_paths():
    assert get_external_module_paths([Path(""), Path("")]) == [Path("")]


def test_load_paths_with_env():
    with patch.object(os, "environ", {"DISSECT_PLUGINS": ":"}):
        assert get_external_module_paths([Path(""), Path("")]) == [Path("")]


class MockOSWarpPlugin(OSPlugin):
    __exports__ = ["f6"]  # OS exports f6
    __findable__ = True
    __name__ = "warp"

    def __init__(self):
        pass

    def get_all_records():
        return []

    def f3(self):
        return "F3"

    def f6(self):
        return "F6"


@patch(
    "dissect.target.plugin.plugins",
    return_value=[
        {"module": "test.x13", "exports": ["f3"], "namespace": "Warp", "class": "x13", "is_osplugin": False},
        {"module": "os", "exports": ["f3"], "namespace": None, "class": "f3", "is_osplugin": False},
        {"module": "os.warp._os", "exports": ["f6"], "namespace": None, "class": "warp", "is_osplugin": True},
    ],
)
@patch("dissect.target.Target", create=True)
@patch("dissect.target.plugin.load")
@pytest.mark.parametrize(
    "search, findable, assert_num_found",
    [
        ("*", True, 3),  # Found with tree search using wildcard
        ("*", False, 0),  # Unfindable plugins are not found...
        ("test.x13.*", True, 1),  # Found with tree search using wildcard, expands to test.x13.f3
        ("test.x13.*", False, 0),  # Unfindable plugins are not found...
        ("test.x13", True, 1),  # Found with tree search, same as above, because users expect +*
        ("test.*", True, 1),  # Found with tree search
        ("test.[!x]*", True, 0),  # Not Found with tree search, all in test not starting with x (no x13)
        ("test.[!y]*", True, 1),  # Found with tree search, all in test not starting with y (so x13 is ok)
        ("test.???.??", True, 1),  # Found with tree search, using question marks
        ("x13", True, 0),  # Not Found: Part of namespace but no match
        ("Warp.*", True, 0),  # Not Found: Namespace != Module so 0
        ("os.warp._os.f6", True, 1),  # Found, OS-plugins also available under verbose name
        ("f6", True, 1),  # Found with classic search
        ("f6", False, 1),  # Backward compatible: unfindable has no effect on classic search
        ("Warp.f3", True, 1),  # Found with classic style search using namespace + function
        ("Warp.f3", False, 1),  # Backward compatible: unfindable has no effect on classic search
        ("f3", True, 1),  # Found with classic style search using only function
        ("os.*", True, 2),  # Found matching os.f3, os.warp._os.f6
        ("os", True, 0),  # Exception for os, because it can be a 'special' plugin (tree match ignored)
    ],
)
def test_find_plugin_functions(plugin_loader, target, plugins, search, findable, assert_num_found):
    os_plugin = MockOSWarpPlugin
    os_plugin.__findable__ = findable
    target._os_plugin = os_plugin
    plugin_loader.return_value = os_plugin()

    found, _ = find_plugin_functions(target, search)
    assert len(found) == assert_num_found


def test_find_plugin_function_windows(target_win: Target) -> None:
    found, _ = find_plugin_functions(target_win, "services")

    assert len(found) == 1
    assert found[0].name == "services"
    assert found[0].path == "os.windows.services.services"


def test_find_plugin_function_linux(target_linux: Target) -> None:
    found, _ = find_plugin_functions(target_linux, "services")

    assert len(found) == 1
    assert found[0].name == "services"
    assert found[0].path == "os.unix.linux.services.services"


TestRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "application/test",
    [
        ("string", "value"),
    ],
)


class _TestNSPlugin(NamespacePlugin):
    __namespace__ = "NS"

    @export(record=TestRecord)
    def test_all(self):
        # Iterate all functions of all subclasses
        yield from self.test()


class _TestSubPlugin1(_TestNSPlugin):
    __namespace__ = "t1"

    @export(record=TestRecord)
    def test(self):
        yield TestRecord(value="test1")


class _TestSubPlugin2(_TestNSPlugin):
    __namespace__ = "t2"

    @export(record=TestRecord)
    def test(self):
        yield TestRecord(value="test2")


class _TestSubPlugin3(_TestSubPlugin2):
    __namespace__ = "t3"

    # Override the test() function of t2
    @export(record=TestRecord)
    def test(self):
        yield TestRecord(value=self._value())

    def _value(self):
        return "test3"


class _TestSubPlugin4(_TestSubPlugin3):
    __namespace__ = "t4"

    # Do not override the test() function of t3, but change the _value function instead.
    def _value(self):
        return "test4"

    @export(record=TestRecord)
    def test_all(self):
        yield TestRecord(value="overridden")


def test_namespace_plugin(target_win: Target) -> None:
    assert "SUBPLUGINS" in dir(_TestNSPlugin)
    # Rename the test functions to protect them from being filtered by NS

    target_win._register_plugin_functions(_TestSubPlugin1(target_win))
    target_win._register_plugin_functions(_TestSubPlugin2(target_win))
    target_win._register_plugin_functions(_TestSubPlugin3(target_win))
    target_win._register_plugin_functions(_TestSubPlugin4(target_win))
    target_win._register_plugin_functions(_TestNSPlugin(target_win))
    assert len(list(target_win.NS.test())) == 4
    assert len(target_win.NS.SUBPLUGINS) == 4

    assert sorted([item.value for item in target_win.NS.test()]) == ["test1", "test2", "test3", "test4"]
    assert sorted([item.value for item in target_win.t1.test()]) == ["test1"]
    assert sorted([item.value for item in target_win.t2.test()]) == ["test2"]
    assert sorted([item.value for item in target_win.t3.test()]) == ["test3"]
    assert sorted([item.value for item in target_win.t4.test()]) == ["test4"]

    # Check whether we can access all subclass functions from the superclass
    assert sorted([item.value for item in target_win.NS.test_all()]) == ["test1", "test2", "test3", "test4"]

    # Check whether we can access the overridden function when explicitly accessing the subplugin
    assert next(target_win.t4.test_all()).value == "overridden"

    # Remove test plugin from list afterwards to avoid order effects
    del PLUGINS["tests"]


def test_find_plugin_function_default(target_default: Target) -> None:
    found, _ = find_plugin_functions(target_default, "services")

    assert len(found) == 2
    names = [item.name for item in found]
    assert "services" in names
    assert "services" in names
    paths = [item.path for item in found]
    assert "os.unix.linux.services.services" in paths
    assert "os.windows.services.services" in paths

    found, _ = find_plugin_functions(target_default, "mcafee.msc")
    assert found[0].path == "apps.av.mcafee.msc"


@pytest.mark.parametrize(
    "pattern",
    [
        ("version,ips,hostname"),
        ("ips,version,hostname"),
        ("hostname,ips,version"),
        ("users,osinfo"),
        ("osinfo,users"),
    ],
)
def test_find_plugin_function_order(target_win: Target, pattern: str) -> None:
    found = ",".join(reduce(lambda rs, el: rs + [el.method_name], find_plugin_functions(target_win, pattern)[0], []))
    assert found == pattern


class _TestIncompatiblePlugin(Plugin):
    def check_compatible(self):
        raise UnsupportedPluginError("My incompatible plugin error")


def test_incompatible_plugin(target_bare: Target) -> None:
    with pytest.raises(UnsupportedPluginError, match="My incompatible plugin error"):
        target_bare.add_plugin(_TestIncompatiblePlugin)


MOCK_PLUGINS = {
    "apps": {  # Plugin descriptors in this branch should be returned for any osfilter
        "mail": {"module": "apps.mail", "functions": "mail"},
    },
    "os": {
        # The OSPlugin for Generic OS, plugins in this branch should only be
        # returned when the osfilter starts with "os." or is None.
        # The _os plugin itself should only be returned if special_keys
        # contains the "_os" key.
        "_os": {"module": "os._os", "functions": "GenericOS"},
        "apps": {
            "app1": {"module": "os.apps.app1", "functions": "app1"},
            "app2": {"module": "os.apps.app2", "functions": "app2"},
        },
        "fooos": {
            # The OSPlugin for FooOS, plugins in this branch should only be
            # returned when the osfilter is "os.fooos" or "os.fooos._os" or
            # None.
            "_os": {"module": "os.foos._os", "functions": "FooOS"},
            "foobar": {"module": "os.foos.foobar", "functions": "foobar"},
            # The plugins under _misc should only be returned if special_keys
            # contains the "_misc" key.
            "_misc": {
                "bar": {"module": "os.foos._misc.bar", "functions": "bar"},
                "tender": {"module": "os.foos._misc.tender", "functions": "tender"},
            },
            "apps": {
                "foo_app": {"module": "os.foos.apps.foo_app", "functions": "foo_app"},
                "bar_app": {"module": "os.foos.apps.bar_app", "functions": "bar_app"},
            },
        },
    },
}


@pytest.mark.parametrize(
    "osfilter, special_keys, only_special_keys, expected_plugin_functions",
    [
        (
            None,
            set(["_os", "_misc"]),
            False,
            [
                "mail",
                "GenericOS",
                "app1",
                "app2",
                "FooOS",
                "foobar",
                "bar",
                "tender",
                "foo_app",
                "bar_app",
            ],
        ),
        (
            "os._os",
            set(["_os"]),
            False,
            [
                "mail",
                "GenericOS",
                "app1",
                "app2",
            ],
        ),
        (
            "os.fooos._os",
            set(),
            False,
            [
                "mail",
                "app1",
                "app2",
                "foobar",
                "foo_app",
                "bar_app",
            ],
        ),
        (
            "os.fooos",
            set(["_os"]),
            False,
            [
                "mail",
                "app1",
                "app2",
                "FooOS",
                "foobar",
                "foo_app",
                "bar_app",
            ],
        ),
        (
            "os.fooos._os",
            set(["_os", "_misc"]),
            True,
            [
                "FooOS",
                "bar",
                "tender",
            ],
        ),
        (
            "bar",
            set(["_os"]),
            False,
            [
                "mail",
            ],
        ),
    ],
)
def test_plugins(
    osfilter: str,
    special_keys: set[str],
    only_special_keys: bool,
    expected_plugin_functions: list[str],
) -> None:
    with (
        patch("dissect.target.plugin.PLUGINS", MOCK_PLUGINS),
        patch("dissect.target.plugin._modulepath", return_value=osfilter),
    ):
        if osfilter is not None:
            # osfilter must be a class or None
            osfilter = Mock

        plugin_descriptors = plugins(
            osfilter=osfilter,
            special_keys=special_keys,
            only_special_keys=only_special_keys,
        )

        plugin_functions = [descriptor["functions"] for descriptor in plugin_descriptors]

        assert sorted(plugin_functions) == sorted(expected_plugin_functions)


def test_plugins_default_plugin(target_default: Target) -> None:
    all_plugins = list(plugins())
    default_plugin_plugins = list(plugins(osfilter=target_default._os_plugin))

    assert default_plugin_plugins == all_plugins

    # The all_with_home is a sentinel function, which should be loaded for a
    # target with DefaultPlugin as OS plugin.
    sentinel_function = "all_with_home"
    has_sentinel_function = False
    for plugin in default_plugin_plugins:
        if sentinel_function in plugin.get("functions", []):
            has_sentinel_function = True
            break

    assert has_sentinel_function

    default_os_plugin_desc = plugins(
        osfilter=target_default._os_plugin,
        special_keys=set(["_os"]),
        only_special_keys=True,
    )

    assert len(list(default_os_plugin_desc)) == 1


@pytest.mark.parametrize(
    "method_name",
    [
        "hostname",
        "ips",
        "version",
        "os",
        "architecture",
    ],
)
def test_os_plugin_property_methods(target_bare: Target, method_name: str) -> None:
    os_plugin = OSPlugin(target_bare)
    with pytest.raises(NotImplementedError):
        getattr(os_plugin, method_name)


class MockOS1(OSPlugin):
    @export(property=True)
    def hostname(self) -> Optional[str]:
        pass

    @export(property=True)
    def ips(self) -> list[str]:
        pass

    @export(property=True)
    def version(self) -> Optional[str]:
        pass

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        pass

    @export(property=True)
    def os(self) -> str:
        pass

    @export(property=True)
    def architecture(self) -> Optional[str]:
        pass


class MockOS2(OSPlugin):
    @export(property=True)
    def hostname(self) -> Optional[str]:
        """Test docstring hostname"""
        pass

    @export(property=True)
    def ips(self) -> list[str]:
        """Test docstring ips"""
        pass

    @export(property=True)
    def version(self) -> Optional[str]:
        """Test docstring version"""
        pass

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        """Test docstring users"""
        pass

    @export(property=True)
    def os(self) -> str:
        """Test docstring os"""
        pass

    @export(property=True)
    def architecture(self) -> Optional[str]:
        """Test docstring architecture"""
        pass


@pytest.mark.parametrize(
    "subclass, replaced",
    [
        (MockOS1, True),
        (MockOS2, False),
    ],
)
def test_os_plugin___init_subclass__(subclass: type[OSPlugin], replaced: bool) -> None:
    exported_methods = {
        "hostname",
        "ips",
        "version",
        "users",
        "os",
        "architecture",
    }

    for method_name in exported_methods:
        os_method = getattr(OSPlugin, method_name)
        if isinstance(os_method, property):
            os_method = os_method.fget
        os_docstring = os_method.__doc__

        subclass_method = getattr(subclass, method_name)
        if isinstance(subclass_method, property):
            subclass_method = subclass_method.fget
        subclass_docstring = subclass_method.__doc__

        assert (os_docstring == subclass_docstring) is replaced
        if not replaced:
            assert subclass_docstring == f"Test docstring {method_name}"


class _TestInternalPlugin(InternalPlugin):
    def test(self) -> None:
        pass


def test_internal_plugin() -> None:
    assert "test" not in _TestInternalPlugin.__exports__
    assert "test" in _TestInternalPlugin.__functions__


class _TestInternalNamespacePlugin(InternalNamespacePlugin):
    __namespace__ = "NS"

    def test(self) -> None:
        pass


def test_internal_namespace_plugin() -> None:
    assert "SUBPLUGINS" in dir(_TestInternalNamespacePlugin)
    assert "test" not in _TestInternalNamespacePlugin.__exports__
    assert "test" in _TestInternalNamespacePlugin.__functions__


class ExampleFooPlugin(Plugin):
    """Example Foo Plugin."""

    def check_compatible(self) -> None:
        return

    @export
    @alias("bar")
    @alias(name="baz")
    def foo(self) -> Iterator[str]:
        """Yield foo!"""
        yield "foo!"


def test_plugin_alias(target_bare: Target) -> None:
    """test ``@alias`` decorator behaviour"""
    target_bare.add_plugin(ExampleFooPlugin)
    assert target_bare.has_function("foo")
    assert target_bare.foo.__aliases__ == ["baz", "bar"]
    assert target_bare.has_function("bar")
    assert target_bare.has_function("baz")
    assert list(target_bare.foo()) == list(target_bare.bar()) == list(target_bare.baz())


@pytest.mark.parametrize(
    "func_path, func",
    [(func.path, func) for func in find_plugin_functions(Target(), "*", compatibility=False, show_hidden=True)[0]],
)
def test_exported_plugin_format(func_path: str, func: PluginFunction) -> None:
    """This test checks plugin style guide conformity for all exported plugins.

    Resources:
        - https://docs.dissect.tools/en/latest/contributing/style-guide.html
    """

    # Ignore DefaultPlugin and NamespacePlugin instances
    if func.class_object.__base__ is NamespacePlugin or func.class_object is DefaultPlugin:
        return

    # Plugin method should specify what it returns
    assert func.output_type in ["record", "yield", "default", "none"], f"Invalid output_type for function {func}"

    py_func = getattr(func.class_object, func.method_name)
    annotations = None

    if hasattr(py_func, "__annotations__"):
        annotations = py_func.__annotations__

    elif isinstance(py_func, property):
        annotations = py_func.fget.__annotations__

    # Plugin method should have a return annotation
    assert annotations and "return" in annotations.keys(), f"No return type annotation for function {func}"

    # TODO: Check if the annotations make sense with the provided output_type

    # Plugin method should have a docstring
    method_doc_str = py_func.__doc__
    assert isinstance(method_doc_str, str), f"No docstring for function {func}"
    assert method_doc_str != "", f"Empty docstring for function {func}"

    # The method docstring should compile to rst without warnings
    assert_valid_rst(method_doc_str)

    # Plugin class should have a docstring
    class_doc_str = func.class_object.__doc__
    assert isinstance(class_doc_str, str), f"No docstring for class {func.class_object.__name__}"
    assert class_doc_str != "", f"Empty docstring for class {func.class_object.__name__}"

    # The class docstring should compile to rst without warnings
    assert_valid_rst(class_doc_str)


def assert_valid_rst(src: str) -> None:
    """Attempts to compile the given string to rst."""

    try:
        publish_string(src, settings_overrides={"halt_level": 2})

    except SystemMessage as e:
        # Limited context was provided to docutils, so some exceptions could incorrectly be raised.
        # We can assume that if the rst is truly invalid this will also be caught by `tox -e build-docs`.
        if "Unknown interpreted text role" not in str(e):
            assert str(e) in src  # makes reading pytest error easier



from unittest.mock import Mock, patch

import pytest

from dissect.target.loaders.phobos import (
    EXTFS_NEEDLE,
    EXTFS_NEEDLE_OFFSET,
    NTFS_NEEDLE,
    PhobosLoader,
)


@pytest.mark.parametrize(
    "needle, offset, expected_offset, element_count, element_size",
    (
        (
            EXTFS_NEEDLE,
            EXTFS_NEEDLE_OFFSET + 1337,
            1337,
            1024,
            1024,
        ),
        (
            NTFS_NEEDLE,
            1337,
            1337,
            1024,
            1024,
        ),
    ),
)
def test_phobos_loader_map(
    needle: bytes,
    offset: int,
    expected_offset: int,
    element_count: int,
    element_size: int,
):
    mock_target = Mock()
    phobos_loader = PhobosLoader(mock_target)
    mock_fh = Mock()
    mock_fs = Mock()
    mock_fs.ntfs.boot_sector.NumberSectors = element_count
    mock_fs.ntfs.sector_size = element_size
    mock_fs.extfs.block_count = element_count
    mock_fs.extfs.block_size = element_size
    expected_size = element_count * element_size

    with patch.object(phobos_loader.path, "open", return_value=mock_fh):
        with (
            patch(
                "dissect.target.loaders.phobos.scrape_pos",
                return_value=[(needle, offset)],
                autospec=True,
            ),
            patch("dissect.util.stream.RelativeStream", autospec=True) as mock_stream,
            patch(
                "dissect.target.filesystem.open",
                return_value=mock_fs,
                autospec=True,
            ),
        ):
            phobos_loader.map(mock_target)
            mock_stream.assert_called_with(mock_fh, expected_offset)
            mock_target.filesystems.add.assert_called_with(mock_fs)
            mock_target.fs.mount.assert_called_with("fs0", mock_fs)
            mock_fh.seek.assert_called_with(offset + expected_size)



import pytest

from dissect.target import Target
from dissect.target.loaders.tar import TarLoader
from dissect.target.plugins.os.windows._os import WindowsPlugin
from tests._utils import absolute_path


def test_tar_loader_compressed_tar_file(target_win: Target) -> None:
    archive_path = absolute_path("_data/loaders/tar/test-archive.tar.gz")

    loader = TarLoader(archive_path)
    loader.map(target_win)

    assert len(target_win.filesystems) == 2

    test_file = target_win.fs.path("test-data/test-file.txt")

    assert test_file.exists()
    assert test_file.open().read() == b"test-value\n"


def test_tar_sensitive_drive_letter(target_bare: Target) -> None:
    tar_file = absolute_path("_data/loaders/tar/uppercase_driveletter.tar")

    loader = TarLoader(tar_file)
    loader.map(target_bare)

    # mounts = / and c:
    assert sorted(target_bare.fs.mounts.keys()) == ["/", "c:"]
    assert "C:" not in target_bare.fs.mounts.keys()

    # Initialize our own WindowsPlugin to override the detection
    target_bare._os_plugin = WindowsPlugin.create(target_bare, target_bare.fs.mounts["c:"])
    target_bare._init_os()

    # sysvol is now added
    assert sorted(target_bare.fs.mounts.keys()) == ["/", "c:", "sysvol"]

    # WindowsPlugin sets the case sensitivity to False
    assert target_bare.fs.get("C:/test.file").open().read() == b"hello_world"
    assert target_bare.fs.get("c:/test.file").open().read() == b"hello_world"


def test_tar_loader_compressed_tar_file_with_empty_dir(target_unix: Target) -> None:
    archive_path = absolute_path("_data/loaders/tar/test-archive-empty-folder.tgz")
    loader = TarLoader(archive_path)
    loader.map(target_unix)

    assert len(target_unix.filesystems) == 2
    test_file = target_unix.fs.path("test/test_file_with_content")
    assert test_file.exists()
    assert test_file.is_file()
    assert test_file.open().read() == b"This is a test!\n"
    empty_folder = target_unix.fs.path("test/empty_dir")
    assert empty_folder.exists()
    assert empty_folder.is_dir()


@pytest.mark.parametrize(
    "archive, expected_drive_letter",
    [
        ("_data/loaders/tar/test-windows-sysvol-absolute.tar", "c:"),  # C: due to backwards compatibility
        ("_data/loaders/tar/test-windows-sysvol-relative.tar", "c:"),  # C: due to backwards compatibility
        ("_data/loaders/tar/test-windows-fs-c-relative.tar", "c:"),
        ("_data/loaders/tar/test-windows-fs-c-absolute.tar", "c:"),
        ("_data/loaders/tar/test-windows-fs-x.tar", "x:"),
    ],
)
def test_tar_loader_windows_sysvol_formats(target_default: Target, archive: str, expected_drive_letter: str) -> None:
    loader = TarLoader(absolute_path(archive))
    loader.map(target_default)

    assert WindowsPlugin.detect(target_default)
    # NOTE: for the sysvol archives, this also tests the backwards compatibility
    assert sorted(target_default.fs.mounts.keys()) == [expected_drive_letter]


def test_tar_anonymous_filesystems(target_default: Target) -> None:
    tar_file = absolute_path("_data/loaders/tar/test-anon-filesystems.tar")

    loader = TarLoader(tar_file)
    loader.map(target_default)

    # mounts = $fs$/fs0, $fs$/fs1 and /
    assert len(target_default.fs.mounts) == 3
    assert "$fs$/fs0" in target_default.fs.mounts.keys()
    assert "$fs$/fs1" in target_default.fs.mounts.keys()
    assert "/" in target_default.fs.mounts.keys()
    assert target_default.fs.get("$fs$/fs0/foo").open().read() == b"hello world\n"
    assert target_default.fs.get("$fs$/fs1/bar").open().read() == b"hello world\n"



import csv
import gzip
from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

AcquireOpenHandlesRecord = TargetRecordDescriptor(
    "filesystem/acquire_open_handles",
    [
        ("path", "name"),
        ("string", "handle_type"),
        ("string", "object"),
        ("varint", "unique_process_id"),
        ("varint", "handle_value"),
        ("varint", "granted_access"),
        ("varint", "creator_back_trace_index"),
        ("varint", "object_type_index"),
        ("varint", "handle_attributes"),
        ("varint", "reserved"),
    ],
)


class OpenHandlesPlugin(Plugin):
    """Plugin to return open file handles collected by Acquire."""

    def __init__(self, target):
        super().__init__(target)
        self.open_handles_file = target.fs.path("$metadata$/open_handles.csv.gz")

    def check_compatible(self) -> None:
        if not self.open_handles_file.exists():
            raise UnsupportedPluginError("No open handles found")

    @export(record=AcquireOpenHandlesRecord)
    def acquire_handles(self) -> Iterator[AcquireOpenHandlesRecord]:
        """Return open handles collected by Acquire.

        An Acquire file container contains an open handles csv when the handles module was used. The content of this csv
        file is returned.
        """
        with self.open_handles_file.open() as fh:
            for row in csv.DictReader(gzip.open(fh, "rt")):
                if name := row.get("name"):
                    row.update({"name": self.target.fs.path(name)})
                yield AcquireOpenHandlesRecord(_target=self.target, **row)



from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.plugins.filesystem.icat import ICatPlugin
from tests._utils import absolute_path


def test_unsupported_fs(target_win):
    # must not fail if sysvol mounted FS is not supported
    assert target_win.icat(1, 0, "") is None


def test_ntfs_fs(target_win):
    mft_data_file = absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw")
    filesystem = NtfsFilesystem(mft=open(mft_data_file, "rb"))
    target_win.filesystems = [filesystem]
    target_win.add_plugin(ICatPlugin)

    assert target_win.icat(1, 0, "") is None



from pathlib import Path
from typing import Any, Iterator

from dissect.esedb.exceptions import Error
from dissect.esedb.tools import ual

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

ClientAccessRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/client_access",
    [
        ("datetime", "last_access_date"),
        ("datetime", "access_date"),
        ("datetime", "insert_date"),
        ("net.ipaddress", "address"),
        ("string", "authenticated_user"),
        ("string", "client_name"),
        ("varint", "access_count"),
        ("varint", "total_access_count"),
        ("string", "tenant_id"),
        ("string", "role_guid"),
        ("string", "role_name"),
        ("string", "product_name"),
        ("string", "path"),
    ],
)

RoleAccessRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/role_access",
    [
        ("datetime", "first_seen_date"),
        ("datetime", "last_seen_date"),
        ("string", "role_guid"),
        ("string", "role_name"),
        ("string", "product_name"),
        ("string", "path"),
    ],
)

VirtualMachineRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/virtual_machines",
    [
        ("datetime", "creation_time"),
        ("datetime", "last_seen_active_date"),
        ("string", "vm_guid"),
        ("string", "bios_guid"),
        ("string", "serial_number"),
        ("string", "path"),
    ],
)

DomainSeenRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/virtual_machines",
    [
        ("datetime", "last_seen_date"),
        ("net.ipaddress", "address"),
        ("string", "resolved_hostname"),
        ("string", "path"),
    ],
)

SystemIdentityRecord = TargetRecordDescriptor(
    "filesystem/windows/ual/system_identity",
    [
        ("datetime", "creation_time"),
        ("varint", "physical_processor_count"),
        ("varint", "cores_per_physical_processor"),
        ("varint", "logical_processors_per_physical_processor"),
        ("varint", "maximum_memory"),
        ("varint", "os_major_version"),
        ("varint", "os_minor_version"),
        ("varint", "os_build"),
        ("varint", "os_platform_id"),
        ("varint", "service_pack_major_version"),
        ("varint", "service_pack_minor_version"),
        ("varint", "os_suite_mask"),
        ("varint", "os_product_type"),
        ("varint", "os_current_time_zone"),
        ("varint", "os_daylight_in_effect"),
        ("string", "system_manufacturer"),
        ("string", "system_product_name"),
        ("string", "system_sm_bios_uuid"),
        ("string", "system_serial_number"),
        ("string", "system_dns_hostname"),
        ("string", "system_domain_name"),
        ("string", "os_serial_number"),
        ("string", "os_country_code"),
        ("string", "os_last_boot_up_time"),
        ("string", "path"),
    ],
)

FIELD_NAME_MAP = {
    "Address": "address",
    "AuthenticatedUserName": "authenticated_user",
    "BIOSGuid": "bios_guid",
    "ClientName": "client_name",
    "CoresPerPhysicalProcessor": "cores_per_physical_processor",
    "CreationTime": "creation_time",
    "FirstSeen": "first_seen_date",
    "HostName": "resolved_hostname",
    "InsertDate": "insert_date",
    "LastAccess": "last_access_date",
    "LastSeen": "last_seen_date",
    "LastSeenActive": "last_seen_active_date",
    "LogicalProcessorsPerPhysicalProcessor": "logical_processors_per_physical_processor",
    "MaximumMemory": "maximum_memory",
    "OSBuildNumber": "os_build",
    "OSCountryCode": "os_country_code",
    "OSCurrentTimeZone": "os_current_time_zone",
    "OSDaylightInEffect": "os_daylight_in_effect",
    "OSLastBootUpTime": "os_last_boot_up_time",
    "OSMajor": "os_major_version",
    "OSMinor": "os_minor_version",
    "OSPlatformId": "os_platform_id",
    "OSProductType": "os_product_type",
    "OSSerialNumber": "os_serial_number",
    "OSSuiteMask": "os_suite_mask",
    "PhysicalProcessorCount": "physical_processor_count",
    "ProductName": "product_name",
    "RoleGuid": "role_guid",
    "RoleName": "role_name",
    "SerialNumber": "serial_number",
    "ServicePackMajor": "service_pack_major_version",
    "ServicePackMinor": "service_pack_minor_version",
    "SystemDNSHostName": "system_dns_hostname",
    "SystemDomainName": "system_domain_name",
    "SystemManufacturer": "system_manufacturer",
    "SystemProductName": "system_product_name",
    "SystemSMBIOSUUID": "system_sm_bios_uuid",
    "SystemSerialNumber": "system_serial_number",
    "TenantId": "tenant_id",
    "TotalAccesses": "total_access_count",
    "VmGuid": "vm_guid",
}


class UalPlugin(Plugin):
    """Return all available User Access Log information.

    User Access Logging (UAL) is a logging system that aggregates client usage data by role and products on a local
    server. It helps Windows server administrators to quantify requests from client computers for roles and services on
    a local server.

    References:
        - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2012-R2-and-2012/hh849634%28v=ws.11%29
    """  # noqa: E501

    __namespace__ = "ual"

    LOG_DB_GLOB = "sysvol/Windows/System32/LogFiles/Sum/*.mdb"

    IDENTITY_DB_FILENAME = "SystemIdentity.mdb"
    IDENTITY_DB_PATH = f"sysvol/Windows/System32/LogFiles/Sum/{IDENTITY_DB_FILENAME}"

    def __init__(self, target):
        super().__init__(target)

        self.mdb_paths = self.find_mdb_files()

        self.role_guid_map = {}
        self.identity_db_parser = None
        self.populate_role_guid_map()

    def check_compatible(self) -> None:
        if not any([path.exists() for path in self.mdb_paths]):
            raise UnsupportedPluginError("No MDB files found")

    def find_mdb_files(self) -> list[Path]:
        return [
            path
            for path in self.target.fs.path("/").glob(self.LOG_DB_GLOB)
            if path.exists() and path.name != self.IDENTITY_DB_FILENAME
        ]

    def populate_role_guid_map(self) -> None:
        identity_db = self.target.fs.path(self.IDENTITY_DB_PATH)
        if not identity_db.exists():
            return

        fh = identity_db.open()
        try:
            self.identity_db_parser = ual.UAL(fh)
        except Error as e:
            self.target.log.warning("Error opening UAL SystemIdentity.mdb database", exc_info=e)
            return

        self.target.log.debug("SystemIdentity.mdb DB loaded")

        for record in self.identity_db_parser.get_table_records("ROLE_IDS"):
            self.role_guid_map[record.get("RoleGuid")] = {
                "product_name": record.get("ProductName"),
                "role_name": record.get("RoleName"),
            }

    def read_table_records(self, table_name: str) -> Iterator[tuple[Path, dict[str, Any]]]:
        for mdb_path in self.mdb_paths:
            fh = mdb_path.open()
            try:
                parser = ual.UAL(fh)
            except Error as e:
                self.target.log.warning("Error opening database: %s", mdb_path, exc_info=e)
                continue

            for table_record in parser.get_table_records(table_name):
                values = {FIELD_NAME_MAP.get(key, key): value for key, value in table_record.items()}
                yield mdb_path, values

    @export(record=ClientAccessRecord)
    def client_access(self) -> Iterator[ClientAccessRecord]:
        """Return client access data within the User Access Logs."""
        for path, client_record in self.read_table_records("CLIENTS"):
            common_values = {k: v for k, v in client_record.items() if k != "activity_counts"}
            role_guid_data = self.role_guid_map.get(common_values.get("role_guid"), {})

            for access_date, access_count in client_record.get("activity_counts", []):
                yield ClientAccessRecord(
                    role_name=role_guid_data.get("role_name"),
                    product_name=role_guid_data.get("product_name"),
                    access_date=access_date,
                    access_count=access_count,
                    path=path,
                    _target=self.target,
                    **common_values,
                )

    @export(record=RoleAccessRecord)
    def role_access(self) -> Iterator[RoleAccessRecord]:
        """Return role access data within the User Access Logs."""
        for path, record in self.read_table_records("ROLE_ACCESS"):
            role_guid_data = self.role_guid_map.get(record.get("role_guid"), {})
            yield RoleAccessRecord(
                role_name=role_guid_data.get("role_name"),
                product_name=role_guid_data.get("product_name"),
                path=path,
                _target=self.target,
                **record,
            )

    @export(record=VirtualMachineRecord)
    def virtual_machines(self) -> Iterator[VirtualMachineRecord]:
        """Return virtual machine data within the User Access Logs."""
        for path, record in self.read_table_records("VIRTUALMACHINES"):
            yield VirtualMachineRecord(path=path, _target=self.target, **record)

    @export(record=DomainSeenRecord)
    def domains_seen(self) -> Iterator[DomainSeenRecord]:
        """Return DNS data within the User Access Logs."""
        for path, record in self.read_table_records("DNS"):
            yield DomainSeenRecord(path=path, _target=self.target, **record)

    @export(record=SystemIdentityRecord)
    def system_identities(self) -> Iterator[SystemIdentityRecord]:
        """Return system identity data within the User Access Logs."""
        if not self.identity_db_parser:
            return
        for record in self.identity_db_parser.get_table_records("SYSTEM_IDENTITY"):
            values = {FIELD_NAME_MAP.get(key, key): value for key, value in record.items()}
            yield SystemIdentityRecord(
                path=self.target.fs.path(self.IDENTITY_DB_PATH),
                _target=self.target,
                **values,
            )



from dissect.target.helpers.record import TargetRecordDescriptor

DefenderMPLogProcessImageRecord = TargetRecordDescriptor(
    "windows/defender/mplog/processimage",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "process_image_name"),
        ("varint", "pid"),
        ("varint", "total_time"),
        ("varint", "count"),
        ("varint", "max_time"),
        ("string", "max_time_file"),
        ("varint", "estimated_impact"),
    ],
)

DefenderMPLogMinFilUSSRecord = TargetRecordDescriptor(
    "windows/defender/mplog/minfiluss",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("path", "path"),
        ("string", "process"),
        ("string", "status"),
        ("string", "state"),
        ("string", "scan_request"),
        ("string", "file_id"),
        ("string", "reason"),
        ("string", "io_status_block_for_new_file"),
        ("string", "desired_access"),
        ("string", "file_attributes"),
        ("string", "scan_attributes"),
        ("string", "access_state_flags"),
        ("string", "backing_file_info"),
    ],
)

DefenderMPLogMinFilBlockedFileRecord = TargetRecordDescriptor(
    "windows/defender/mplog/blockedfile",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "blocked_file"),
        ("string", "process"),
        ("string", "status"),
        ("string", "state"),
        ("string", "scan_request"),
        ("string", "file_id"),
        ("string", "reason"),
        ("string", "io_status_block_for_new_file"),
        ("string", "desired_access"),
        ("string", "file_attributes"),
        ("string", "scan_attributes"),
        ("string", "access_state_flags"),
        ("string", "backing_file_info"),
    ],
)


DefenderMPLogBMTelemetryRecord = TargetRecordDescriptor(
    "windows/defender/mplog/bmtelemetry",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "guid"),
        ("varint", "signature_id"),
        ("string", "sigsha"),
        ("varint", "threat_level"),
        ("varint", "process_id"),
        ("varint", "process_creation_time"),
        ("varint", "session_id"),
        ("path", "image_path"),
        ("string", "taint_info"),
        ("string", "operations"),
    ],
)

DefenderMPLogEMSRecord = TargetRecordDescriptor(
    "windows/defender/mplog/ems",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "process"),
        ("varint", "pid"),
        ("string", "sigseq"),
        ("varint", "send_memory_scan_report"),
        ("varint", "source"),
    ],
)

DefenderMPLogOriginalFileNameRecord = TargetRecordDescriptor(
    "windows/defender/mplog/originalfilename",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "original_file_name"),
        ("path", "full_path"),
        ("string", "hr"),
    ],
)

DefenderMPLogExclusionRecord = TargetRecordDescriptor(
    "windows/defender/mplog/exclusion",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("path", "full_path_with_drive_letter"),
        ("path", "full_path_with_device_path"),
    ],
)

DefenderMPLogLowfiRecord = TargetRecordDescriptor(
    "windows/defender/mplog/lowfi",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("command", "lowfi"),
    ],
)

DefenderMPLogDetectionAddRecord = TargetRecordDescriptor(
    "windows/defender/mplog/detectionadd",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "detection"),
    ],
)


DefenderMPLogThreatRecord = TargetRecordDescriptor(
    "windows/defender/mplog/threat",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("command", "threat"),
    ],
)

DefenderMPLogDetectionEventRecord = TargetRecordDescriptor(
    "windows/defender/mplog/detectionevent",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "threat_type"),
        ("command", "command"),
    ],
)

DefenderMPLogResourceScanRecord = TargetRecordDescriptor(
    "windows/defender/mplog/resourcescan",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string", "scan_id"),
        ("varint", "scan_source"),
        ("datetime", "start_time"),
        ("datetime", "end_time"),
        ("string", "resource_schema"),
        ("path", "resource_path"),
        ("varint", "result_count"),
        ("string[]", "threats"),
        ("path[]", "resources"),
    ],
)

DefenderMPLogThreatActionRecord = TargetRecordDescriptor(
    "windows/defender/mplog/threataction",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("string[]", "threats"),
        ("path[]", "resources"),
        ("string[]", "actions"),
    ],
)

DefenderMPLogRTPRecord = TargetRecordDescriptor(
    "windows/defender/mplog/rtp_log",
    [
        ("datetime", "ts"),
        ("path", "source_log"),
        ("datetime", "last_perf"),
        ("datetime", "first_rtp_scan"),
        ("string", "plugin_states"),
        ("path[]", "process_exclusions"),
        ("path[]", "path_exclusions"),
        ("string[]", "ext_exclusions"),
    ],
)



import pytest

from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.plugins.filesystem.ntfs.usnjrnl import UsnjrnlPlugin
from tests._utils import absolute_path


@pytest.fixture
def mocked_timeline_plugin(target_win):
    filesystem = NtfsFilesystem(usnjrnl=open(absolute_path("_data/plugins/filesystem/ntfs/usnjrnl/usnjrnl.bin"), "rb"))
    target_win.filesystems = [filesystem]
    plugin = UsnjrnlPlugin(target_win)
    return plugin


def test_usnjrnl_normal(mocked_timeline_plugin):
    """Test parsing of a usnjrnl file"""

    data = list(mocked_timeline_plugin.usnjrnl())

    assert len(data) == 15214



import io
from urllib.parse import urlparse

import pytest
from flow.record.fieldtypes import path as flow_path

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loaders.log import LogLoader


@pytest.mark.parametrize(
    "path, uri, input_file, expected_mapping",
    [
        ("/dir/*.evt*", None, "/dir/test.evtx", "/sysvol/windows/system32/winevt/logs/test.evtx"),
        ("/dir/*.evt*", None, "/dir/test.evt", "/sysvol/windows/system32/config/test.evt"),
        ("/source/iis.log", "log:///dir/with/files/*.log?hint=iis", "/source/iis.log", "/sysvol/files/logs/iis.log"),
    ],
)
def test_log_loader(target_default: Target, path: str, uri: str, input_file: str, expected_mapping: str) -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_fh(input_file, io.BytesIO(b"\x00"))
    log_loader = LogLoader(vfs.path(path), parsed_path=urlparse(uri))
    log_loader.map(target_default)
    # TODO: RGLOB does not take into account the seperator of the target, so maybe an issue with the flavour?
    observed_mapping = next(target_default.fs.path("/").rglob("*.*"))
    assert expected_mapping == flow_path(observed_mapping)



from __future__ import annotations

import logging
import socket
import ssl
import time
import urllib
from io import DEFAULT_BUFFER_SIZE
from pathlib import Path
from struct import pack, unpack
from typing import Optional, Union

from dissect.util.stream import AlignedStream

from dissect.target.containers.raw import RawContainer
from dissect.target.exceptions import LoaderError
from dissect.target.loader import Loader
from dissect.target.plugin import arg
from dissect.target.target import Target

log = logging.getLogger(__name__)


class RemoteStream(AlignedStream):
    def __init__(self, stream: RemoteStreamConnection, disk_id: int, size: Optional[int] = None):
        self.stream = stream
        self.disk_id = disk_id
        super().__init__(size)

    def _read(self, offset: int, length: int) -> bytes:
        return self.stream.read(self.disk_id, offset, length)


class RemoteStreamConnection:
    # Max. number of times we try to reconnect (still tweaking this)
    MAX_RECONNECTS = 30

    # Max. number of short reads, short reads might happen because of internal bugs
    # this is a mechanism to make sure the client will not just hang
    MAX_SHORT_READS = 10

    # Time to wait before attempting to reconnect
    RECONNECT_WAIT = 10

    # Socket timeout, connections can be slow, so not too low
    # Also there might be an initial waiting time of 10s because of the
    # previous connection! So it must be at least 10s.
    SOCKET_TIMEOUT = 30

    # Remote agent understands 3 commands:
    # 1      INFO: return disk size and sector size for each remote disk
    # 2      QUIT: stops the agent on the remote machine
    # 50 + X READ: read disk number X (starts with 0)
    COMMAND_INFO = 1
    COMMAND_QUIT = 2
    COMMAND_READ = 50

    # This values can be injected by other tooling to let the Remote Loader
    # take advantage of embedded SSL artefacts.
    CONFIG_KEY = None
    CONFIG_CRT = None

    @staticmethod
    def configure(key, crt):
        RemoteStreamConnection.CONFIG_KEY = key
        RemoteStreamConnection.CONFIG_CRT = crt

    def __init__(self, hostname: str, port: int, **kwargs):
        self.hostname = hostname
        self.port = port
        self._is_connected = False
        self._socket = None
        self._ssl_sock = None
        self._context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self._context.verify_mode = ssl.CERT_REQUIRED
        self._context.load_default_certs()
        self._max_reconnects = self.MAX_RECONNECTS
        self._max_shortreads = self.MAX_SHORT_READS
        self._reconnect_wait = self.RECONNECT_WAIT
        self._socket_timeout = self.SOCKET_TIMEOUT

        flag_cert_chain_loaded = False
        flag_verify_locations_loaded = False

        if options := kwargs.get("options"):
            client_key = options.get("key")
            client_crt = options.get("crt")
            server_ca = options.get("ca")
            noverify = options.get("noverify")

            if client_key and client_crt:
                self._context.load_cert_chain(certfile=client_crt, keyfile=client_key)
                flag_cert_chain_loaded = True
            if noverify:
                self._context.verify_mode = ssl.CERT_NONE
            if server_ca:
                self._context.load_verify_locations(server_ca)
                flag_verify_locations_loaded = True
            self._max_reconnects = options.get("reconnects", max(0, self._max_reconnects))
            self._max_shortreads = options.get("shortreads", max(0, self._max_shortreads))
            self._reconnect_wait = options.get("reconnectwait", max(0, self._reconnect_wait))
            self._socket_timeout = options.get("sockettimeout", max(0, self._socket_timeout))

        if flag_cert_chain_loaded is False and self.CONFIG_KEY is not None and self.CONFIG_CRT is not None:
            self._context.load_cert_chain_str(certfile=self.CONFIG_CRT, keyfile=self.CONFIG_KEY)

        if flag_verify_locations_loaded is False and self.CONFIG_CRT is not None:
            self._context.load_verify_locations(cadata=self.CONFIG_CRT)

        self.log = log

    def is_connected(self) -> bool:
        return self._is_connected

    def connect(self) -> None:
        if self._is_connected:
            return

        reconnects = 0
        while self._is_connected is False:
            self.log.debug("Connecting to agent")
            # Even during the handshake things can go wrong with unreliable connections
            try:
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self._socket.settimeout(self._socket_timeout)
                ssl_sock = self._context.wrap_socket(self._socket, server_hostname=self.hostname)
                ssl_sock.connect((self.hostname, self.port))
                self._ssl_sock = ssl_sock
                self._is_connected = True
                self.log.debug("Connection established with agent")
            except Exception:
                # If the max. connections are exceeded it is probably no use anymore because
                # the remote system has become unreachable and we need to report back to the
                # user that we can no longer contact the remote machine.
                if reconnects > self._max_reconnects:
                    raise ConnectionError("Unable to reconnect with remote agent.")
                if self._ssl_sock is not None:
                    self._ssl_sock.close()
                if self._socket is not None:
                    self._socket.close()
                # Directly re-connecting seem to be less succesful, allow some time to re-connect
                # seems to yield best results in practice
                self.log.debug("Unable to connect to agent, next attempt in %d sec.", self._reconnect_wait)
                time.sleep(self._reconnect_wait)
                reconnects += 1

    def _receive_bytes(self, length: int) -> bytes:
        data = b""
        received = 0
        short_reads = 0
        while received < length:
            packet = self._ssl_sock.recv(min(length - received, DEFAULT_BUFFER_SIZE))
            packet_size = len(packet)
            data += packet
            received += packet_size
            if packet_size == 0:
                short_reads += 1
            if short_reads > self._max_shortreads:
                raise TimeoutError("Too many short reads.")

        return data

    def read(self, disk_id: int, offset: int, length: int) -> bytes:
        # In practice this should not occur but if one uses this stream directly we will
        # be kind and inform the user about the limitations of this logic
        if length < 0:
            raise NotImplementedError("RemoteStreamConnection does not support length = -1")

        if length == 0:
            return b""

        data = b""
        received = 0
        while received < length:
            self.connect()
            self._ssl_sock.send(pack(">BQQ", self.COMMAND_READ + disk_id, offset, length))
            try:
                data += self._receive_bytes(length)
                received += length
            except Exception:
                self.log.debug("Unable to read data from agent, re-connecting")
                self._ssl_sock.close()
                self._socket.close()
                self._is_connected = False

        return data

    def close(self) -> None:
        if self.is_connected:
            self._ssl_sock.send(pack(">BQQ", self.COMMAND_QUIT, 0, 0))

    def info(self) -> list[RemoteStream]:
        self.connect()
        response_size = 16
        self._ssl_sock.send(pack(">BQQ", self.COMMAND_INFO, 0, 0))
        number_of_disks = unpack("<B", self._ssl_sock.recv(1))[0]
        remainder = number_of_disks * response_size
        data = self._receive_bytes(remainder)
        disks = []
        for i in range(0, number_of_disks * response_size, response_size):
            part = data[i : i + response_size]
            (disk_size, _) = unpack("<QQ", part)
            disks.append(RemoteStream(self, i // response_size, disk_size))

        return disks


@arg("--remote-key", dest="key", help="private key")
@arg("--remote-crt", dest="crt", help="client certificate")
@arg("--remote-ca", dest="ca", help="certificate Authority")
@arg("--remote-noverify", dest="noverify", help="no certificate verification")
@arg("--remote-reconnects", dest="reconnects", help="max number of reconnects")
@arg("--remote-shortreads", dest="shortreads", help="max limit shortreads")
@arg("--remote-reconnectwait", dest="reconnectwait", help="max time before reconnection attempt")
@arg("--remote-sockettimeout", dest="sockettimeout", help="socket timeout")
class RemoteLoader(Loader):
    """Load a remote target that runs a compatible Dissect agent."""

    def __init__(self, path: Union[Path, str], **kwargs):
        super().__init__(path)
        uri = kwargs.get("parsed_path")
        if uri is None:
            raise LoaderError("No URI connection details has been passed.")
        options = dict(urllib.parse.parse_qsl(uri.query, keep_blank_values=True))
        self.stream = RemoteStreamConnection(uri.hostname, uri.port, options=options)

    def map(self, target: Target) -> None:
        self.stream.log = target.log
        for disk in self.stream.info():
            target.disks.add(RawContainer(disk))

    @staticmethod
    def detect(path: Path) -> bool:
        # You can only activate this loader by URI-scheme "remote://"
        return False



import os
import random
import shutil
from unittest.mock import Mock

import pytest

from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.tools.fs import _extract_path, cp


@pytest.fixture
def vfs(files) -> VirtualFilesystem:
    vfs = VirtualFilesystem()
    for file in files:
        if file[-1] == "/":
            vfs.makedirs(file)
        else:
            vfs.map_file_entry(file, VirtualFile(vfs, file, None))
    return vfs


@pytest.mark.parametrize("files", [["file"]])
def test_extract_file(vfs):
    output_file = f"/tmp/file{random.randint(0, 1000)}"
    try:
        _extract_path(vfs.path("file"), output_file)
    except Exception:  # noqua
        # The files are virtual, so we expect the method to raise an exception
        pass

    assert os.path.exists(output_file)
    # cleanup
    os.remove(output_file)


@pytest.mark.parametrize("files", [[]])
def test_file_not_exist(vfs):
    output_file = f"/tmp/file{random.randint(0, 1000)}"

    _extract_path(vfs.path("file"), output_file)

    assert not os.path.exists(output_file)


@pytest.mark.parametrize("files", [["dir/"]])
def test_extract_directory(vfs):
    output = f"/tmp/dir{random.randint(0, 1000)}"

    _extract_path(vfs.path("dir"), f"{output}/test")

    assert os.path.isdir(f"{output}/test")

    # cleanup
    shutil.rmtree(output)


@pytest.mark.parametrize("files", [["dir/", "dir/test"]])
def test_cp_file_path(vfs):
    args = Mock()
    args.output = f"/tmp/dir{random.randint(0, 1000)}"

    cp(None, vfs.path("dir/test"), args)

    assert os.path.isfile(f"{args.output}/test")

    # cleanup
    shutil.rmtree(args.output)


@pytest.mark.parametrize("files", [["dir/", "dir/test"]])
def test_cp_directory(vfs):
    args = Mock()
    args.output = f"/tmp/dir{random.randint(0, 1000)}"

    cp(None, vfs.path("dir"), args)

    assert os.path.isfile(f"{args.output}/test")

    # cleanup
    shutil.rmtree(args.output)


@pytest.mark.parametrize("files", [[]])
def test_cp_non_existing_file(vfs):
    args = Mock()
    args.output = f"/tmp/dir{random.randint(0, 1000)}"

    cp(None, vfs.path("dir/test"), args)

    assert not os.path.exists(args.output)


@pytest.mark.parametrize(
    "files",
    [["dir/", "dir/test", "dir/subdirectory_1/", "dir/subdirectory_2/", "dir/subdirectory_3/subdirectory_4/"]],
)
def test_cp_subdirectories(vfs, files):
    args = Mock()
    args.output = f"/tmp/dir{random.randint(0, 1000)}"

    cp(None, vfs.path("dir/"), args)

    filesystem_files = (file.replace("dir/", "") for file in files)

    for directories in filesystem_files:
        assert os.path.exists(f"{args.output}/{directories}")
    # cleanup
    shutil.rmtree(args.output)
